/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
 * This is GNU GO, a Go program. Contact gnugo@gnu.org, or see   *
 * http://www.gnu.org/software/gnugo/ for more information.      *
 *                                                               *
 * Copyright 1999 and 2000 by the Free Software Foundation.      *
 *                                                               *
 * This program is free software; you can redistribute it and/or *
 * modify it under the terms of the GNU General Public License   *
 * as published by the Free Software Foundation - version 2.     *
 *                                                               *
 * This program is distributed in the hope that it will be       *
 * useful, but WITHOUT ANY WARRANTY; without even the implied    *
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR       *
 * PURPOSE.  See the GNU General Public License in file COPYING  *
 * for more details.                                             *
 *                                                               *
 * You should have received a copy of the GNU General Public     *
 * License along with this program; if not, write to the Free    *
 * Software Foundation, Inc., 59 Temple Place - Suite 330,       *
 * Boston, MA 02111, USA                                         *
\* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <assert.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "liberty.h"
#include "cache.h"


/*
 * The functions in reading.c are used to read whether groups 
 * can be captured or not. See the Texinfo documentation 
 * (Reading) for more information.
 *
 * NULL POINTERS: Many functions in this file can use pointers
 * to return the locations of recommended plays. These can be
 * set NULL in which case these values are not returned.
 */

static int do_find_defense(int si, int sj, int *i, int *j, int komaster);
static int defend1(int si, int sj, int *i, int *j, int komaster);
static int defend2(int si, int sj, int *i, int *j, int komaster);
static int defend3(int si, int sj, int *i, int *j, int komaster);
static int defend4(int si, int sj, int *i, int *j, int komaster);
static int special_rescue(int si, int sj, int ai, int aj, 
			  int *ti, int *tj, int komaster);
static int special_rescue2(int si, int sj, int libi[2], int libj[2], 
			   int *ti, int *tj, int komaster);
static int do_attack(int si, int sj, int *i, int *j, int komaster);
static int attack1(int si, int sj, int *i, int *j, int komaster);
static int attack2(int si, int sj, int *i, int *j, int komaster);
static int attack3(int si, int sj, int *i, int *j, int komaster);
static int attack4(int si, int sj, int *i, int *j, int komaster);
static int find_cap2(int si, int sj, int *i, int *j, int komaster);
static int find_cap3(int si, int sj, int *i, int *j, int komaster);
static int special_attack2(int si, int sj, int libi[2], int libj[2], 
			   int *ti, int *tj, int komaster);
static int edge_attack(int si, int sj, int ai, int aj, int *ki, int *kj);
static int edge_defend(int si, int sj, int ai, int aj, int *ki, int *kj);
static int break_chain(int si, int sj, int *i, int *j, int *k, 
		       int *l, int komaster);
static int break_chain2(int si, int sj, int *i, int *j, int komaster);
static int break_chain3(int si, int sj, int *i, int *j, int komaster);
static int double_atari_chain2(int si, int sj, int *i, int *j);
static int relative_break_chain(int, int, int *, int *, int, int);
static void order_moves(int si, int sj, int num_moves, int *movei, int *movej,
			int color);
static int naive_ladder_defense(int si, int sj, int ai, int aj,
				int bi, int bj, int color, int other);
static int naive_ladder_break_through(int si, int sj, int ai, int aj,
				      int color, int other);

/* Statistics. */
static int reading_node_counter = 0;



/* ================================================================ */  
/*                          Goal functions                          */
/* ================================================================ */


/*
 * These functions define goals for the reading process.  They use 
 * the rest of the reading machinery to evaluate wether the goal
 * is fullfillable.
 *
 * The simplest goals are defined by attack() and find_defense(),
 * namely to see if it is possible to capture or defend a single
 * string.  More complex goals are defined by e.g. attack_either()
 * or defend_both().
 *
 * The functions in this section are the only ones which are 
 * callable from outside this file.
 */


/* attack(si, sj, *i, *j) determines if the string at (m, n) can be 
 * captured, and if so, (*i, *j) returns the attacking move, unless
 * (*i, *j) are null pointers. Use null pointers if you are interested
 * in the result of the attack but not the attacking move itself.
 *
 * Return 1 if the attack succeeds unconditionally, 0 if it doesn't.
 * Returns 2 or 3 if the result depends on ko: 
 *   - Returns 2 if the attack succeeds provided attacker is willing to
 *     ignore any ko threat (the attacker makes the first ko capture).
 *   - Returns 3 if attack succeeds provided attacker has a ko threat
 *     which must be answered (the defender makes the first ko capture).
 */

int
attack(int si, int sj, int *i, int *j)
{
  return do_attack(si, sj, i, j, EMPTY);
}


/* find_defense(m, n, *i, *j) attempts to find a move that will save
 * the string at (m,n). It returns 1 if such a move is found, with
 * (*i, *j) the location of the saving move, unless (*i, *j) are
 * null pointers. It is not checked that tenuki defends, so this may 
 * give an erroneous answer if !attack(m,n).
 * 
 * Returns 2 or 3 if the result depends on ko. Returns 2 if the
 * string can be defended provided the defender is willing to ignore
 * any ko threat. Returns 3 if the defender wins by having a ko threat
 * which must be answered.
 */

int 
find_defense(int si, int sj, int *i, int *j)
{
  return do_find_defense(si, sj, i, j, EMPTY);
}


/*
 * attack_either(ai, aj, bi, bj) returns true if there is a move which
 * guarantees that at least one of the strings (ai, aj) and (bi, bj)
 * can be captured. A typical application for this is in connection
 * patterns, where after a cut it suffices to capture one of the cutting
 * stones.
 *
 * FIXME: The current implementation only looks for uncoordinated
 *        attacks. This is insufficient to find double ataris or 
 *        moves such as 'a' in positions like
 *
 *        XOOOOOOOX
 *        XOXXOXXOX
 *        XX..a..XX
 *        ---------
 *
 *        where neither of the threatened X stones can be captured right
 *        out. Still either can be captured by a move down to a.
 */

int
attack_either(int ai, int aj, int bi, int bj)
{
  int color = p[ai][aj];
  ASSERT(color != EMPTY , ai, aj);
  ASSERT(color == p[bi][bj], bi, bj);

  /* Start by attacking the string with the fewest liberties. On
   * average this seems to be slightly more efficient.
   */
  if (countlib(ai, aj) <= countlib(bi, bj))
    return attack(ai, aj, NULL, NULL) || attack(bi, bj, NULL, NULL);
  else
    return attack(bi, bj, NULL, NULL) || attack(ai, aj, NULL, NULL);
}


/*
 * defend_both(ai, aj, bi, bj) returns true if both the strings (ai, aj)
 * and (bi, bj) can be defended simultaneously or if there is no attack.
 * A typical application for this is in connection patterns, where
 * after a cut it's necessary to defend both cutting stones.
 *
 * FIXME: The current implementation only looks at uncoordinated defense.
 *        A proper implementation would require some serious reading.
 */

int
defend_both(int ai, int aj, int bi, int bj)
{
  int a_threatened = 0;
  int b_threatened = 0;
  int ci, cj, di, dj;
  
  int color = p[ai][aj];
  ASSERT(color != EMPTY , ai, aj);
  ASSERT(color == p[bi][bj], bi, bj);

  if (attack(ai, aj, NULL, NULL)) {
    a_threatened = 1;
    if (!find_defense(ai, aj, &ci, &cj))
      return 0; /* (ai, aj) already lost */
  }
  
  if (attack(bi, bj, NULL, NULL)) {
    b_threatened = 1;
    if (!find_defense(bi, bj, &di, &dj))
      return 0; /* (bi, bj) already lost */
  }

  /* Neither string can be attacked or only one of them, in which case
   * we have time to save it.
   */
  if (!a_threatened || !b_threatened)
    return 1;
  
  /* If both strings are threatened we assume that one will become lost,
   * unless find_defense() happened to return the same defense point for
   * both (which e.g. may happen if they are in fact the same string).
   * This is still a bit too pessimistic, as there may be one move which
   * saves both strings. To do this right we should try each move which
   * defends either string and see if it also defends the other string.
   */

  if (ci == di && cj == dj)
    return 1; /* Both strings can be attacked but also defended by one move. */

  /* We also try each of the returned defense points and see whether
   * the other string can still be attacked. This still gives a
   * somewhat pessimistic estimation.
   */

  if (trymove(ci, cj, color, "defend_both-A", ai, aj)) {
    if (p[bi][bj] && !attack(bi, bj, NULL, NULL)) {
      popgo();
      return 1;
    }
    popgo();
  }
  
  if (trymove(di, dj, color, "defend_both-B", bi, bj)) {
    if (p[ai][aj] && !attack(ai, aj, NULL, NULL)) {
      popgo();
      return 1;
    }
    popgo();
  }
  
  /* Both strings can be attacked but we have only time to defend one. */
  return 0;
}


/*
 * break_through(ai, aj, bi, bj, ci, cj) returns 1 if a position can
 * succesfully be broken through and 2 if it can be cut. The position
 * is assumed to have the shape (the colors may be reversed)
 *
 * .O.       dbe
 * OXO       aFc
 *
 * It is X to move and try to capture at least one of a, b, and c. If
 * this succeeds, X is said to have broken through the position.
 * Otherwise X may try to cut through the position, which means
 * keeping F safe and getting a tactically safe string at either d or
 * e.
 *
 * Important notice: a, b, and c must be given in the correct order.
 *
 * FIXME: The reading involved here can most likely be improved.
 *
 * FIXME: We need to take ko results properly into account.
 */

static int
break_through_helper(int ai, int aj, int bi, int bj, int ci, int cj,
		     int di, int dj, int ei, int ej, int Fi, int Fj,
		     int color, int other);

int
break_through(int ai, int aj, int bi, int bj, int ci, int cj)
{
  int di, dj;
  int ei, ej;
  int Fi, Fj;
  int gi, gj;
  
  int color = p[ai][aj];
  int other = OTHER_COLOR(color);

  int success = 0;
  int success2 = 0;
  
  /* Basic sanity checking. */
  ASSERT(color != EMPTY , ai, aj);
  ASSERT(color == p[bi][bj], bi, bj);
  ASSERT(color == p[ci][cj], ci, cj);

  /* Construct the rest of the points in the pattern. */
  Fi = (ai + ci) / 2;    /* F midpoint between a and c. */
  Fj = (aj + cj) / 2;
  di = ai + bi - Fi;     /* Use diagonal relation a+b = d+F. */
  dj = aj + bj - Fj;
  ei = bi + ci - Fi;     /* Use diagonal relation b+c = e+F. */
  ej = bj + cj - Fj;

  /* More sanity checking. */
  ASSERT(p[di][dj] == EMPTY , di, dj);
  ASSERT(p[ei][ej] == EMPTY , ei, ej);

  /* F might already have been captured. (play_break_through_n() can't
   * detect this.
   */
  if (p[Fi][Fj] == EMPTY)
    return 0;
  
  ASSERT(p[Fi][Fj] == other , Fi, Fj);

  /* First X tries to play at d. */
  success = break_through_helper(ai, aj, bi, bj, ci, cj,
				 di, dj, ei, ej, Fi, Fj, color, other);
  if (success == 1)
    return 1;
  
  success2 = break_through_helper(ci, cj, bi, bj, ai, aj,
				  ei, ej, di, dj, Fi, Fj, color, other);

  if (success2 == 1)
    return 1;

  if (success2 == 2)
    success = 2;

  /* If we haven't been lucky yet, we might need to start by
   * defending F.
   *
   * FIXME: The function would probably be considerably faster if we
   * start by checking whether F needs defense. Beware of ko potential
   * though.
   */
  success2 = 0;
  if (attack(Fi, Fj, NULL, NULL) && find_defense(Fi, Fj, &gi, &gj)) {
    if (trymove(gi, gj, other, "break_through-A", Fi, Fj)) {
      /* Now we let O defend his position by playing either d or e.
       * FIXME: There may be other plausible moves too.
       */
      if (trymove(di, dj, color, "break_through-B", Fi, Fj)) {
	/* O connects at d, so X cuts at e. */
	if (safe_move(ei, ej, other)) {
	  success2 = 2;
	  if (!p[ci][cj] || attack(ci, cj, NULL, NULL))
	    success2 = 1;
	}
	popgo();
      }

      if (success2 > 0 && trymove(ei, ej, color, "break_through-C", Fi, Fj)) {
	/* O connects at e, so X cuts at d. */
	if (safe_move(di, dj, other)) {
	  /* success2 is already 1 or 2. */
	  if (p[ai][aj] && !attack(ai, aj, NULL, NULL))
	    success2 = 2;
	}
	else
	  success2 = 0;
	popgo();
      }
      popgo();
    }
  }
    
  if (success2 >= 1)
    return success2;

  return success;
}

/* Helper function for break_through(). Since we can symmetrically
 * start by cutting at d or e, we use the same code for both attacks,
 * simply switching positions between the two calls.
 */
static int
break_through_helper(int ai, int aj, int bi, int bj, int ci, int cj,
		     int di, int dj, int ei, int ej, int Fi, int Fj,
		     int color, int other)
{
  int success = 0;
  int gi, gj;

  if (trymove(di, dj, other, "break_through_helper-A", Fi, Fj)) {
    /* If F can be attacked we can't start in this way. */
    if (!attack(Fi, Fj, NULL, NULL)) {
      /* If d is safe too, we have at least managed to break through. */
      if (!attack(di, dj, &gi, &gj)) {
	success = 2;
	/* So now we can try to capture something. */
	if (!p[ai][aj] || !p[bi][bj] || !defend_both(ai, aj, bi, bj))
	  success = 1;
	else {
	  /* Both a and b could be defended, or didn't need to be.
	   * Let's see if a move at e is sufficient for O.
	   */
	  int attack_on_b = 0;
	  int attack_on_a = 0;

	  if (trymove(ei, ej, color, "break_through_helper-B", Fi, Fj)) {
	    if (attack(bi, bj, NULL, NULL))
	      attack_on_b = 1;
	    else if (attack(ai, aj, NULL, NULL))
	      attack_on_a = 1;
	    popgo();
	  }

	  /* Let O find a defense and play it. */
	  if (attack_on_a || attack_on_b) {
	    int hi = -1;
	    int hj = -1;
	    if (attack_on_b)
	      find_defense(ai, aj, &hi, &hj);
	    else
	      find_defense(bi, bj, &hi, &hj);
	    
	    if (hi != -1
		&& trymove(hi, hj, color, "break_through_helper-C", Fi, Fj)) {
	      /* Now we make a second cut at e, trying to capture
	       * either b or c.
	       */
	      if (trymove(ei, ej, other, "break_through_helper-D", Fi, Fj)) {
		if (!p[bi][bj] || !p[ci][cj] || !defend_both(bi, bj, ci, cj))
		  success = 1;
		popgo();
	      }
	      popgo();
	    }
	    else
	      success = 1; /* This should have been covered by
			    * defend_both(), so probably unnecessary. */
	  }
	}
      }

      /* Too bad, d could be attacked. We let O play the attack and
       * then try to make a second cut at e. But first we must test if
       * O at e is sufficient to capture d.
       */
      else {
	if (trymove(ei, ej, color, "break_through_helper-E", Fi, Fj)) {
	  if (!p[di][dj] || !find_defense(di, dj, NULL, NULL)) {
	    popgo();
	    popgo();
	    return 0;
	  }
	  popgo();
	}
	
	if (gi == ei && gj == ej) {
	  popgo();
	  return 0;
	}

	if (trymove(gi, gj, color, "break_through_helper-F", Fi, Fj)) {
	  if (trymove(ei, ej, other, "break_through_helper-G", Fi, Fj)) {
	    if (!attack(ei, ej, NULL, NULL)) {
	      success = 2;
	      if (!p[bi][bj] || !p[ci][cj] || !defend_both(bi, bj, ci, cj))
		success = 1;
	    }
	    popgo();
	  }
	  popgo();
	}
      }
	
    }
    popgo();
  }

  return success;
}


/* Like find_defense, but takes the komaster argument. If the
 * opponent is komaster, reading functions will not try
 * to take ko.
 */

static int
do_find_defense(int si, int sj, int *i, int *j, int komaster)
{
  int xi, xj;
  int can_save;
  int mylib;
  int             found_read_result;
  Read_result   * read_result;
  
  SETUP_TRACE_INFO("find_defense", si, sj);
  
  RTRACE("Can we rescue %m?\n", si, sj);

  /* We first check if the number of liberties is larger than four. In
   * that case we don't cache the result and to avoid needlessly
   * storing the position in the hash table, we must do this test
   * before we look for cached results.
   */
  mylib = countlib(si, sj);
  if (mylib>4)   /* no need to cache result in this case */
    return 1;

  if ((stackp <= depth) && (hashflags & HASH_FIND_DEFENSE)) {
    found_read_result = get_read_result(FIND_DEFENSE, &si, &sj, &read_result);
    if (found_read_result) {
      TRACE_CACHED_RESULT(*read_result);
      if (rr_get_result(*read_result) != 0) {
	if (i) *i = rr_get_result_i(*read_result);
	if (j) *j = rr_get_result_j(*read_result);
      }

      SGFTRACE(rr_get_result_i(*read_result), rr_get_result_j(*read_result),
	       rr_get_result(*read_result), "cached");
      return rr_get_result(*read_result);
    }

    /* This data should always be recorded. */
    if (read_result) {
      rr_set_routine_i_j_stackp(*read_result, FIND_DEFENSE, si, sj, stackp);
    }
  } else
    read_result = NULL;

  if (mylib == 1) {
    can_save = defend1(si, sj, &xi, &xj, komaster);
    READ_RETURN(read_result, i, j, xi, xj, can_save);
  }
  else if (mylib == 2) {
    can_save = defend2(si, sj, &xi, &xj, komaster);
    if (can_save) {
      RTRACE("saving move for %m found at %m!\n", si, sj, xi, xj);
      READ_RETURN(read_result, i, j, xi, xj, can_save);
    }
  }
  else if (mylib == 3) {
    if (stackp > depth) 
      return 1;
    if (defend3(si, sj, &xi, &xj, komaster)) {
      RTRACE("saving move for %m found at %m!\n", si, sj, xi, xj);
      READ_RETURN(read_result, i, j, xi, xj, 1);
    }
  }
  else if (mylib == 4) {
    if (stackp > depth) 
      return 1;
    if (defend4(si, sj, &xi, &xj, komaster)) {
      RTRACE("saving move for %m found at %m!\n", si, sj, xi, xj);
      READ_RETURN(read_result, i, j, xi, xj, 1);
    }
  }

  READ_RETURN0(read_result);
}


/* If si, sj points to a string with exactly one liberty, defend1 
 * determines whether it can be saved by extending or capturing
 * a boundary chain having one liberty. The function returns 1 if the string
 * can be saved, otherwise 0. It returns 2 or 3 if it can be saved, conditioned
 * on ko. Returns 2 if it can be saved provided (color) is willing to
 * ignore any ko threat. Returns 3 if it can be saved if (color) has
 * a ko threat which must be answered.
 *
 * The pair defend1-attack2 call each other recursively to
 * read situations such as ladders. They read all ladders to the end.
 * If the reading ply (stackp) is deeper than the deep-reading cutoff
 * parameter depth, whose default value DEPTH is defined in liberty, then a
 * string is assumed alive if it can get 3 liberties. When 
 * fourlib_depth < stackp < depth, a string is considered alive if it can get
 * four liberties. When stackp < fourlib_depth, it is considered alive
 * if it can get 5 liberties.
 *
 */

static int 
defend1(int si, int sj, int *i, int *j, int komaster)
{
  int color;
  int xi, xj;
  int bcode;
  int savei = -1, savej = -1;
  int saveresult = 0;
  int liberties;
  int can_catch;
  int             found_read_result;
  Read_result   * read_result;

  SETUP_TRACE_INFO("defend1", si, sj);
  reading_node_counter++;
  
  assert(p[si][sj] != EMPTY);
  if (countlib(si, sj) != 1) {
    dump_stack();
    abort();
  }

  RTRACE("try to escape atari on %m.\n",  si, sj);

  color = p[si][sj];

  if ((stackp <= depth) && (hashflags & HASH_DEFEND1)) {
  
    found_read_result = get_read_result(DEFEND1, &si, &sj, &read_result);
    if (found_read_result) {
      TRACE_CACHED_RESULT(*read_result);
      if (rr_get_result(*read_result) != 0) {
	if (i) *i = rr_get_result_i(*read_result);
	if (j) *j = rr_get_result_j(*read_result);
      }

      SGFTRACE(rr_get_result_i(*read_result), rr_get_result_j(*read_result),
	       rr_get_result(*read_result), "cached");
      return rr_get_result(*read_result);
    }

    /* This data should always be recorded. */
    if (read_result) {
      rr_set_routine_i_j_stackp(*read_result, DEFEND1, si, sj, stackp);
    }
  } else
    read_result = NULL;

  /* If we can capture a bounding chain and live, then do so. */
  bcode = break_chain(si, sj, &xi, &xj, NULL, NULL, komaster);
  if (bcode) {
    if (bcode != 1) {
      savei = xi;
      savej = xj;
      saveresult = bcode;
    }
    else {
      SGFTRACE(xi, xj, 1, "break chain");
      READ_RETURN(read_result, i, j, xi, xj, 1);
    }
  }
  
  /* (xi, xj) will be the liberty of the string. */
  liberties = findlib(si, sj, 1, &xi, &xj);
  ASSERT(liberties == 1, si, sj);

  /* Try to extend along our own liberty. */
  RTRACE("extending to %m.\n", xi, xj);
  if (trymove(xi, xj, color, "defend1-A", si, sj)) {
    liberties = countlib(xi, xj);

    /* If we are still in atari, or have no liberties at all, we lose. */
    if (liberties < 2) {
      if (savei == -1) {
	RTRACE("still atari: it dies!\n");
	popgo();
	SGFTRACE(-1, -1, 0, "extension ineffective - A");
	READ_RETURN0(read_result);
      }
      else {
	RTRACE("saved by ko at %m\n", savei, savej);
	popgo();
	SGFTRACE(savei, savej, saveresult, "saved move - A");
	READ_RETURN(read_result, i, j, savei, savej, saveresult);
      }
    }
    
    /* If we are still within the deep reading cutoff, then read further
     * even if the group has 3 liberties. 
     */
    if ((stackp <= depth) && (liberties==3)) {

      /* If the group can be killed in spite of the 3 liberties, return 0. */
      if (attack3(si, sj, NULL, NULL, komaster)) {
	if (savei == -1) {
	  RTRACE("it dies!\n");
	  popgo();
	  SGFTRACE(-1, -1, 0, "extension ineffective - B");
	  READ_RETURN0(read_result);
	}
	else {
	  RTRACE("saved by ko at %m\n", savei, savej);
	  popgo();
	  SGFTRACE(savei, savej, saveresult, "saved move - B");
	  READ_RETURN(read_result, i, j, savei, savej, saveresult);
	}
      } else {
	RTRACE("it lives at %m!\n", xi, xj);
	popgo();
	SGFTRACE(xi, xj, 1, "extension effective - A");
	READ_RETURN(read_result, i, j, xi, xj, 1);
      }
    }
    else if ((stackp <= fourlib_depth) && (liberties==4)) {
      if (attack4(si, sj, NULL, NULL, komaster)) {
	if (savei == -1) {
	  RTRACE("it dies!\n");
	  popgo();
	  SGFTRACE(-1, -1, 0, "extension ineffective - C");
	  READ_RETURN0(read_result);
	}
	else {
	  RTRACE("saved by ko at %m\n", savei, savej);
	  popgo();
	  SGFTRACE(savei, savej, saveresult, "saved move - C");
	  READ_RETURN(read_result, i, j, savei, savej, saveresult);
	}
      } else {
	RTRACE("it lives at %m!\n", xi, xj);
	popgo();
	SGFTRACE(xi, xj, 1, "extension effective - B");
	READ_RETURN(read_result, i, j, xi, xj, 1);
      }
    }

    /* If we are here and liberties is >2 (must be 3), then stackp must
     * be bigger than depth, and then we consider 3 liberties to live.  
     */
    if (liberties>2) {
      RTRACE("it lives at %m!\n", xi, xj);
      popgo();
      SGFTRACE(xi, xj, 1, ">2 liberties lives");
      READ_RETURN(read_result, i, j, xi, xj, 1);
    }
    
    /* If we are here, then liberties must be 2 and we don't care if
       stackp is greater than depth. We continue the ladder anyway. */
    ASSERT(liberties==2, si, sj);
    
    can_catch=attack2(si, sj, NULL, NULL, komaster);
    if (can_catch==1) {
      if (savei == -1) {
	RTRACE("DEAD!\n");
	popgo();
	SGFTRACE(-1, -1, 0, "extension ineffective - D");
	READ_RETURN0(read_result);
      }
      else {
	popgo();
	SGFTRACE(savei, savej, saveresult, "saved move - D");
	READ_RETURN(read_result, i, j, savei, savej, saveresult);
      }
    } else if (can_catch==2) {
      RTRACE("START KO FIGHT at %m!\n", xi, xj);
      popgo();
      SGFTRACE(xi, xj, 3, "extension gives ko - A");
      READ_RETURN(read_result, i, j, xi, xj, 3);
    } else if (can_catch==3) {
      RTRACE("START KO FIGHT at %m!\n", xi, xj);
      popgo();
      SGFTRACE(xi, xj, 2, "extension gives ko - B");
      READ_RETURN(read_result, i, j, xi, xj, 2);
    } else {
      RTRACE("ALIVE at %m!\n", xi, xj);
      popgo();
      SGFTRACE(xi, xj, 1, "extension effective - C");
      READ_RETURN(read_result, i, j, xi, xj, 1);
    }
  }

  /* If we get here, then trymove() failed => the ladder works. */
  RTRACE( "Cannot extend: die in ladder\n");

  if (savei == -1) {
    SGFTRACE(-1, -1, 0, NULL);
    READ_RETURN0(read_result);
  }
  else {
    SGFTRACE(savei, savej, saveresult, "saved move - E");
    READ_RETURN(read_result, i, j, savei, savej, saveresult);
  }
}


/* If si, sj points to a group with two liberties, defend2 determines
 * whether the group can be saved by extending, or by capturing part of
 * its surrounding chain. A group is considered safe if either part of
 * the surrounding chain may be captured, or if it can get 3
 * liberties. It is presumed that the opponent could kill if tenuki.
 * If both extensions work, it prefers the one which maximizes 
 * liberties.
 *
 * i and j return the move to save the stones. Can be NULL if caller
 * is not interested in the details.
*/

static int 
defend2(int si, int sj, int *i, int *j, int komaster)
{
  int color, other, ai, aj, bi, bj, firstlib, secondlib;
  int xi, xj;
  int liberties;
  int libi[2], libj[2];
  int acode = 0, bcode = 0;
  int savei = -1, savej = -1;
  int savecode = 0;
  int val = 0;
  int k;
  int             found_read_result;
  Read_result   * read_result;

  SETUP_TRACE_INFO("defend2", si, sj);
  reading_node_counter++;
  
  RTRACE("trying to rescue %m\n",  si, sj);
  color=p[si][sj];
  other=OTHER_COLOR(color);

  assert(p[si][sj]!=EMPTY);
  assert(countlib(si, sj) == 2);

  if ((stackp <= depth) && (hashflags & HASH_DEFEND2)) {
  
    found_read_result = get_read_result(DEFEND2, &si, &sj, &read_result);
    if (found_read_result) {
      TRACE_CACHED_RESULT(*read_result);
      if (rr_get_result(*read_result) != 0) {
	if (i) *i = rr_get_result_i(*read_result);
	if (j) *j = rr_get_result_j(*read_result);
      }

      SGFTRACE(rr_get_result_i(*read_result), rr_get_result_j(*read_result),
	       rr_get_result(*read_result), "cached");
      return rr_get_result(*read_result);
    }

    /* This data should always be recorded. */
    if (read_result) {
      rr_set_routine_i_j_stackp(*read_result, DEFEND2, si, sj, stackp);
    }
  } else
    read_result = NULL;

  val = break_chain(si, sj, &xi, &xj, NULL, NULL, komaster);
  if (val == 1) {
    SGFTRACE(xi, xj, 1, "break chain");
    READ_RETURN(read_result, i, j, xi, xj, 1);
  }
  
  if (val >= 2) {
    savei = xi;
    savej = xj;
    savecode = val;
  }
  liberties = findlib(si, sj, 2, libi, libj);
  ASSERT(liberties == 2, si, sj);
  order_moves(si, sj, liberties, libi, libj, color);

  ai = libi[0];
  aj = libj[0];
  bi = libi[1];
  bj = libj[1];

  /* We prefer not to crawl along the edge any longer
   * than necessary, so we try edge_defend first */

  if (edge_defend(si, sj, ai, aj, &xi, &xj))
    READ_RETURN(read_result, i, j, xi, xj, 1);
  if (edge_defend(si, sj, bi, bj, &xi, &xj))
    READ_RETURN(read_result, i, j, xi, xj, 1);

  RTRACE("trying extension to %m\n", ai, aj);
  if (trymove(ai, aj, color, "defend2-A", si, sj)) {
    acode = 0;

    firstlib = countlib(si, sj);
    if (firstlib == 2)
      acode = attack2(si, sj, NULL, NULL, komaster);
    else if ((stackp <= depth)
	&& (firstlib == 3))
      acode = attack3(si, sj, NULL, NULL, komaster); 
    else if ((stackp <= fourlib_depth)
	     && (firstlib == 4))
      acode = attack4(si, sj, NULL, NULL, komaster);
    popgo();
    /* if the first move works with ko we save it, then look for
     * something better.
     */
    if (acode == 3) {
      savei = ai;
      savej = aj;
      savecode = 2;
    } 
    else if ((savecode != 2) && (acode == 2)) {
      savei = ai;
      savej = aj;
      savecode = 3;
    }
    if (acode)
      firstlib = 0;

  /* If i is not NULL then the caller wants to know a working move. 
   * In this case we want to return the *best* move. If
   * not, we don't care, so if we know the first move works
   * we do not check the second move.
   */

    if (firstlib > 1 && !i) {
      RTRACE("%m rescues\n", ai, aj);
      SGFTRACE(ai, aj, 1, "extension rescues - A");
      READ_RETURN(read_result, i, j, ai, aj, 1);
    }
    else
      RTRACE("%m does not rescue. Is other move better?\n", ai, aj);
  } else {
    RTRACE("%m isn't a legal move!\n", ai, aj);
    firstlib=0;
  }

  RTRACE("trying extension to %m\n", bi, bj);
  if (trymove(bi, bj, color, "defend2-B", si, sj)) {
    secondlib = countlib(si, sj);
    if ((firstlib >= secondlib) && (firstlib > 2)) {
      RTRACE("%m is not superior to first move\n",  bi, bj);
      popgo();
      SGFTRACE(ai, aj, 1, "extension rescues - B");
      READ_RETURN(read_result, i, j, ai, aj, 1);
    }
    
    if (secondlib == 2)
      bcode = attack2(si, sj, NULL, NULL, komaster);
    else if ((stackp <= depth)
	     && (secondlib == 3))
      bcode = attack3(si, sj, NULL, NULL, komaster); 
    else if ((stackp <= fourlib_depth)
	     && (secondlib == 4))
      bcode = attack4(si, sj, NULL, NULL, komaster);
    popgo();
    if ((firstlib >= secondlib) && (firstlib > 2)) {
      RTRACE("%m is not superior to first move\n",  bi, bj);
      READ_RETURN(read_result, i, j, ai, aj, 1);
    }
    if ((secondlib > 1) && (secondlib >= firstlib) && (bcode == 0)) {
      if (firstlib > 1)
	RTRACE("%m also rescues, we'll use it\n", bi, bj);
      else 
	RTRACE("%m rescues, we'll use it\n",  bi, bj);
      READ_RETURN(read_result, i, j, bi, bj, 1);
    }
    
    if (bcode == 3) {
      savei = bi;
      savej = bj;
      savecode = 2;
    } 
    else if ((savecode != 2) && (bcode == 2)) {
      savei = bi;
      savej = bj;
      savecode = 3;
    }
    if (bcode == 0 && secondlib > 1) {
      SGFTRACE(bi, bj, 1, "extension rescues - C");
      READ_RETURN(read_result, i, j, bi, bj, 1);
    }
  }
  if (firstlib > 1 && acode == 0) {
    RTRACE("%m doesn't work, use first move\n", bi, bj);
    SGFTRACE(ai, aj, 1, "extension rescues - D");
    READ_RETURN(read_result, i, j, ai, aj, 1);
  }
  
  /* This used to be conditioned on stackp < depth, but let's
   * try it unconditionally for a while. This fixes incident 64.
   */
  if (1) {
    int bc = break_chain2(si, sj, &xi, &xj, komaster);
    if (bc) {
      SGFTRACE(xi, xj, bc, "break chain2");
      READ_RETURN(read_result, i, j, xi, xj, bc);
    }
  }
  else {
    if (double_atari_chain2(si, sj, &xi, &xj)) {
      SGFTRACE(xi, xj, 1, "double atari chain2");
      READ_RETURN(read_result, i, j, xi, xj, 1);
    }
  }
  
  if (stackp <= fourlib_depth) {
    int bc = (break_chain3(si, sj, &xi, &xj, komaster));
    if (bc) {
      SGFTRACE(xi, xj, bc, "break chain3");
      READ_RETURN(read_result, i, j, xi, xj, bc);
    }
  }
  
  if (stackp <= depth) {
    for (k = 0; k < liberties; ++k) {
      if (special_rescue(si, sj, libi[k], libj[k], &xi, &xj, komaster)) {
	SGFTRACE(xi, xj, 1, "special rescue");
	READ_RETURN(read_result, i, j, xi, xj, 1);
      }
    }
  }
  
  if (stackp <= backfill_depth) {
    val = special_rescue2(si, sj, libi, libj, &xi, &xj, komaster);
    if (val == 1) {
      SGFTRACE(xi, xj, 1, "special rescue2");
      READ_RETURN(read_result, i, j, xi, xj, 1);
    }
    else if (val > 1) {
      savei = xi;
      savej = xj;
      savecode = val;
    }
  }
  
  if (savei != -1) {
    SGFTRACE(savei, savej, savecode, "saved move");
    READ_RETURN(read_result, i, j, savei, savej, savecode);
  }

  RTRACE("failed to find rescuing move.\n");
  SGFTRACE(-1, -1, 0, NULL);
  READ_RETURN0(read_result);
}


/* defend3(si, sj, *i, *j) attempts to find a move rescuing the 
 * string at (si, sj) with 3 liberties.  If such a move can be found,
 * it returns true, and if the pointers i, j are not NULL, 
 * then it returns the saving move in (*i, *j).
 */

static int 
defend3(int si, int sj, int *i, int *j, int komaster)
{
  int color;
  int xi, xj;
  int liberties;
  int libi[3], libj[3];
  int maxlib;
  int savei=-1, savej=-1;
  int savecode = 0;
  int bc=0;
  int k;
  int             found_read_result;
  Read_result   * read_result;

  SETUP_TRACE_INFO("defend3", si, sj);
  reading_node_counter++;

  RTRACE("trying to rescue %m\n",  si, sj);
  color=p[si][sj];

  assert(p[si][sj]!=EMPTY);
  assert(countlib(si, sj) == 3);

  if ((stackp <= depth) && (hashflags & HASH_DEFEND3)) {
    found_read_result = get_read_result(DEFEND3, &si, &sj, &read_result);
    if (found_read_result) {
      TRACE_CACHED_RESULT(*read_result);
      if (rr_get_result(*read_result) != 0) {
	if (i) *i = rr_get_result_i(*read_result);
	if (j) *j = rr_get_result_j(*read_result);
      }

      SGFTRACE(rr_get_result_i(*read_result), rr_get_result_j(*read_result),
	       rr_get_result(*read_result), "cached");
      return rr_get_result(*read_result);
    }

    /* This data should always be recorded. */
    if (read_result) {
      rr_set_routine_i_j_stackp(*read_result, DEFEND3, si, sj, stackp);
    }
  } else
    read_result = NULL;

  /* If we can capture a surrounding string, this is the best way to defend. */
  bc=break_chain(si, sj, &xi, &xj, NULL, NULL, komaster);
  if (bc == 1) {
    SGFTRACE(xi, xj, 1, "break chain");
    READ_RETURN(read_result, i, j, xi, xj, 1);
  }

  /* We prefer not to sacrifice unnecessarily. If the attacking
   * move works but can itself be captured, we cache its location
   * in (savei, savej). Then if we cannot find a better way, we 
   * use it at the end.
   */
  if (bc >= 2) {
    savei = xi;
    savej = xj;
    savecode = bc;
  }

  /* Get the three liberties into (libi, libj). */
  liberties = findlib(si, sj, 3, libi, libj);
  ASSERT(liberties == 3, si, sj);

  /* We prefer not to crawl along the edge any longer
   * than necessary so we try edge_defend first.
   */

  if (stackp <= depth)
    for (k = 0; k < liberties; ++k)
      if (edge_defend(si, sj, libi[k], libj[k], &xi, &xj))
	READ_RETURN(read_result, i, j, xi, xj, 1);

  order_moves(si, sj, liberties, libi, libj, color);

  /* Before trying seriously, check if there is a simple solution. */
  if (stackp > fourlib_depth)
    maxlib = 4;
  else
    maxlib = 5;
  for (k = 0; k < liberties; ++k)
    if (approxlib(libi[k], libj[k], color, maxlib, NULL, NULL) > maxlib-1) {
      SGFTRACE(libi[k], libj[k], 1, "trivial extension");
      READ_RETURN(read_result, i, j, libi[k], libj[k], 1);
    }

  /* Nope, no quick solution available. Try reading further instead. */
  for (k = 0; k < liberties; ++k) {
    if (stackp >= branch_depth && k > 0)
      break;
    RTRACE("trying extension to %m\n", libi[k], libj[k]);
    if (trymove(libi[k], libj[k], color, "defend3-A", si, sj)) {
      if (!do_attack(si, sj, NULL, NULL, komaster)) {
	popgo();
	SGFTRACE(libi[k], libj[k], 1, "extend");
	READ_RETURN(read_result, i, j, libi[k], libj[k], 1);
      }
      popgo();
    } 
  } 

  if (stackp <= backfill_depth) {
    bc = (break_chain2(si, sj, &xi, &xj, komaster));
    if (bc) {
      SGFTRACE(xi, xj, bc, "break chain2");
      READ_RETURN(read_result, i, j, xi, xj, bc);
    }
  }
  if (stackp <= fourlib_depth) {
    bc = (break_chain3(si, sj, &xi, &xj, komaster));
    if (bc) {
      SGFTRACE(xi, xj, bc, "break chain3");
      READ_RETURN(read_result, i, j, xi, xj, bc);
    }
  }

  if (savei != -1) {
    SGFTRACE(savei, savej, savecode, "saved move");
    READ_RETURN(read_result, i, j, savei, savej, savecode);
  }

  /* If nothing else works, try to defend with second order liberties. */
  if (stackp <= depth) {
    for (k = 0; k < liberties; ++k) {
      if (special_rescue(si, sj, libi[k], libj[k], &xi, &xj, komaster)) {
	SGFTRACE(xi, xj, 1, "special rescue");
	READ_RETURN(read_result, i, j, xi, xj, 1);
      }
    }
  }

  RTRACE("failed to find rescuing move.\n");
  SGFTRACE(-1, -1, 0, NULL);
  READ_RETURN0(read_result);
}


/* defend4(si, sj, *i, *j) attempts to find a move rescuing the 
 * string at (si, sj) with 4 liberties.  If such a move can be found,
 * it returns true, and if the pointers i, j are not NULL, 
 * then it returns the saving move in (*i, *j).
 */

static int 
defend4(int si, int sj, int *i, int *j, int komaster)
{
  int color;
  int xi, xj;
  int liberties;
  int libi[4], libj[4];
  int maxlib;
  int savei=-1, savej=-1;
  int savecode = 0;
  int bc=0;
  int k;
  int             found_read_result;
  Read_result   * read_result;
  
  SETUP_TRACE_INFO("defend4", si, sj);
  reading_node_counter++;

  RTRACE("trying to rescue %m\n",  si, sj);
  color=p[si][sj];

  assert(p[si][sj]!=EMPTY);
  assert(countlib(si, sj) == 4);

  if ((stackp <= depth) && (hashflags & HASH_DEFEND4)) {
    found_read_result = get_read_result(DEFEND4, &si, &sj, &read_result);
    if (found_read_result) {
      TRACE_CACHED_RESULT(*read_result);
      if (rr_get_result(*read_result) != 0) {
	if (i) *i = rr_get_result_i(*read_result);
	if (j) *j = rr_get_result_j(*read_result);
      }

      SGFTRACE(rr_get_result_i(*read_result), rr_get_result_j(*read_result),
	       rr_get_result(*read_result), "cached");
      return rr_get_result(*read_result);
    }

    /* This data should always be recorded. */
    if (read_result) {
      rr_set_routine_i_j_stackp(*read_result, DEFEND4, si, sj, stackp);
    }
  } else
    read_result = NULL;

  /* If we can capture a surrounding string, this is the best way to defend. */
  bc=break_chain(si, sj, &xi, &xj, NULL, NULL, komaster);
  if (bc == 1) {
    SGFTRACE(xi, xj, 1, "break chain");
    READ_RETURN(read_result, i, j, xi, xj, 1);
  }

  /* We prefer not to sacrifice unnecessarily. If the attacking
   * move works but can itself be captured, we cache its location
   * in (savei, savej). Then if we cannot find a better way, we 
   * use it at the end.
   */       
  if (bc >= 2) {
    savei = xi;
    savej = xj;
    savecode = bc;
  }

  liberties = findlib(si, sj, 4, libi, libj);
  ASSERT(liberties == 4, si, sj);
  order_moves(si, sj, liberties, libi, libj, color);

  /* Before trying seriously, check if there is a simple solution. */
  if (stackp > fourlib_depth)
    maxlib = 4;
  else
    maxlib = 5;
  for (k = 0; k < liberties; ++k)
    if (approxlib(libi[k], libj[k], color, maxlib, NULL, NULL) > maxlib-1) {
      SGFTRACE(libi[k], libj[k], 1, "trivial extension");
      READ_RETURN(read_result, i, j, libi[k], libj[k], 1);
    }

  /* Nope, no quick solution available. Try reading further instead. */
  for (k = 0; k < liberties; ++k) {
    if (stackp >= branch_depth && k > 0)
      break;
    RTRACE("trying extension to %m\n", libi[k], libj[k]);
    if (trymove(libi[k], libj[k], color, "defend4-A", si, sj)) {
      if (!do_attack(si, sj, NULL, NULL, komaster)) {
	popgo();
	SGFTRACE(libi[k], libj[k], 1, "extend");
	READ_RETURN(read_result, i, j, libi[k], libj[k], 1);
      }
      popgo();
    } 
  } 

  if ((stackp <= backfill_depth) 
      && break_chain2(si, sj, &xi, &xj, komaster)) {
    SGFTRACE(xi, xj, 1, "break chain2");
    READ_RETURN(read_result, i, j, xi, xj, 1);
  }

  if (savei != -1) {
    SGFTRACE(savei, savej, savecode, "saved move");
    READ_RETURN(read_result, i, j, savei, savej, savecode);
  }

  RTRACE("failed to find rescuing move.\n");
  SGFTRACE(-1, -1, 0, NULL);
  READ_RETURN0(read_result);
}


/*
 * special_rescue(si, sj, ai, aj, *ti, *tj) is called with (si, sj) a
 * string having a liberty at (ai, aj). The saving move is returned
 * in (*ti, *tj).
 *
 * This checks whether a move on a second order liberty is a rescuing
 * move, e.g. in shapes like:
 *
 *   .        O        O       X.XXO
 *  O.*  or  ..*  or  O.*  or  XOOXO
 *   O        O        O       ...*.
 *                             -----
 *
 * This will occasionally save a string where no other move will. To
 * reduce the branching caused by these moves, we require that the
 * opponent can be trivially captured when trying to intercept on the
 * corresponding first order liberty.
 */

static int
special_rescue(int si, int sj, int ai, int aj, int *ti, int *tj, int komaster)
{
  int other = OTHER_COLOR(p[si][sj]);
  int k;

  /* Loop over the four neighbours of the liberty, (ai+di, aj+dj). */
  for (k=0; k<4; k++) {
    int di = deltai[k];
    int dj = deltaj[k];
    if (ON_BOARD(ai+di, aj+dj)
	&& p[ai+di][aj+dj] == EMPTY) {
      if (trymove(ai+di, aj+dj, p[si][sj], "special_rescue", si, sj)) {
	/* Use approxlib() to test for trivial capture. */
	if (approxlib(ai, aj, other, 2, NULL, NULL) <= 1
	    /* FIXME: Take care of return value 2 and 3. */
	    && !do_attack(si, sj, NULL, NULL, komaster)) {
	  popgo();
	  if (ti) *ti = ai+di;
	  if (tj) *tj = aj+dj;
	  return 1;
	}
	popgo();
      }
    }
  }
  
  return 0;
}


/* In a situation like this:
 *       
 *   OOXXXX     the following code can find the
 *   .OXOOX     defensive move at 'c'.
 *   .cO.OX
 *   .X.OOX
 *   ------
 */
static int
special_rescue2(int si, int sj, int libi[2], int libj[2], 
		int *ti, int *tj, int komaster)
{
  int  color=p[si][sj];
  int  other=OTHER_COLOR(color);
  int  savecode = 0;
  int  newlibi[3];
  int  newlibj[3];
  int  xi, xj;
  int  savei = -1, savej = -1;
  int  k;

  for (k = 0; k < 2; ++k) {
    /* Let (ai, aj) and (bi, bj) be the two liberties. Reverse the
     * order during the second pass through the loop.
     */
    int ai = libi[k];
    int aj = libj[k];
    int bi = libi[1-k];
    int bj = libj[1-k];
    if (suicide(ai, aj, other) 
	&& (approxlib(ai, aj, color, 3, newlibi, newlibj)==2)) {
      if ((newlibi[0] != bi) || (newlibj[0] != bj)) {
	xi = newlibi[0];
	xj = newlibj[0];
      } else {
	xi = newlibi[1];
	xj = newlibj[1];
      }

      if (trymove(xi, xj, color, "special_rescue2", si, sj)) {
	int acode = do_attack(si, sj, NULL, NULL, komaster);
	if (acode != 1) {
	  if (acode == 0) {
	    popgo();
	    if (ti) *ti = xi;
	    if (tj) *tj = xj;
	    return 1;
	  }
	  else if (acode == 2)
	    savecode=3;
	  else if (acode == 3)
	    savecode=2;
	  savei=xi;
	  savej=xj;
	}
	popgo();
      }
    }
  }

  if (savecode) {
    if (ti) *ti = savei;
    if (tj) *tj = savej;
    return savecode;
  }

  return 0;
}


/* 
 * If (si,sj) points to a string and (ai,aj) a liberty of it,
 * returns 1 if (ai,aj) is an edge liberty in the following
 * configuration:
 * 
 * Xb?
 * Oa.
 * ---
 *
 * and if a move at 'b' rescues the O string.
 * 
 * FIXME: (1) So far we do not treat ko;
 *
 */

static int
edge_defend(int si, int sj, int ai, int aj, int *ki, int *kj)
{
  int color = p[si][sj];

  if ((ai == 0) 
      && (p[ai+1][aj] == EMPTY)
      && ((aj > 0) && (aj < board_size-1))
      && (((p[ai][aj-1] == color)
	   && (p[ai+1][aj-1] == color)
	   && (p[ai][aj+1] == EMPTY))
	  || ((p[ai][aj+1] == color)
	      && (p[ai+1][aj+1] == color)
	      && (p[ai][aj-1] == EMPTY))))
    if (trymove(ai+1, aj, color, "edge_defend", si, sj)) {
      if (!attack(si, sj, NULL, NULL)) {
	if (ki) *ki = ai+1;
	if (ki) *kj = aj;
	popgo();
	return 1;
      }
      popgo();
    }
  if ((ai == board_size-1) 
      && (p[ai-1][aj] == EMPTY)
      && ((aj > 0) && (aj < board_size-1))
      && (((p[ai][aj-1] == color)
	   && (p[ai-1][aj-1] == color)
	   && (p[ai][aj+1] == EMPTY))
	  || ((p[ai][aj+1] == color)
	      && (p[ai-1][aj+1] == color)
	      && (p[ai][aj-1] == EMPTY))))
    if (trymove(ai-1, aj, color, "edge_defend", si, sj)) {
      if (!attack(si, sj, NULL, NULL)) {
	if (ki) *ki = ai-1;
	if (ki) *kj = aj;
	popgo();
	return 1;
      }
      popgo();
    }
  if ((aj == 0) 
      && (p[ai][aj+1] == EMPTY)
      && ((ai > 0) && (ai < board_size-1))
      && (((p[ai-1][aj] == color)
	   && (p[ai-1][aj+1] == color)
	   && (p[ai+1][aj] == EMPTY))
	  || ((p[ai+1][aj] == color)
	      && (p[ai+1][aj+1] == color)
	      && (p[ai-1][aj] == EMPTY))))
    if (trymove(ai, aj+1, color, "edge_defend", si, sj)) {
      if (!attack(si, sj, NULL, NULL)) {
	if (ki) *ki = ai;
	if (ki) *kj = aj+1;
	popgo();
	return 1;
      }
      popgo();
    }
  if ((aj == board_size-1) 
      && (p[ai][aj-1] == EMPTY)
      && ((ai > 0) && (ai < board_size-1))
      && (((p[ai-1][aj] == color)
	   && (p[ai-1][aj-1] == color)
	   && (p[ai+1][aj] == EMPTY))
	  || ((p[ai+1][aj] == color)
	      && (p[ai+1][aj-1] == color)
	      && (p[ai-1][aj] == EMPTY))))
    if (trymove(ai, aj-1, color, "edge_defend", si, sj)) {
      if (!attack(si, sj, NULL, NULL)) {
	if (ki) *ki = ai;
	if (ki) *kj = aj-1;
	popgo();
	return 1;
      }
      popgo();
    }
  return 0;
}


/* ================================================================ */  
/*                       Aggressive functions                       */
/* ================================================================ */


/* Like attack, but takes the komaster argument. If the
 * opponent is komaster, reading functions will not try
 * to take ko.
 */

static int 
do_attack(int si, int sj, int *i, int *j, int komaster)
{
  int color = p[si][sj];
  int xi, xj;
  int libs;
  int result;
  int             found_read_result;
  Read_result   * read_result;

  SETUP_TRACE_INFO("attack", si, sj);

  ASSERT(color != 0, si, sj);

  if (color == 0)      /* if assertions are turned off, silently fails */
    return 0;

  if ((stackp <= depth) && (hashflags & HASH_ATTACK)) {
    found_read_result = get_read_result(ATTACK, &si, &sj, &read_result);
    if (found_read_result) {
      TRACE_CACHED_RESULT(*read_result);
      if (rr_get_result(*read_result) != 0) {
	if (i) *i = rr_get_result_i(*read_result);
	if (j) *j = rr_get_result_j(*read_result);
      }

      SGFTRACE(rr_get_result_i(*read_result), rr_get_result_j(*read_result),
	       rr_get_result(*read_result), "cached");
      return rr_get_result(*read_result);
    }

    /* This data should always be recorded. */
    if (read_result) {
      rr_set_routine_i_j_stackp(*read_result, ATTACK, si, sj, stackp);
    }
  } else
    read_result = NULL;

  /* Treat the attack differently depending on how many liberties the 
     string at (si, sj) has. */
  libs = countlib(si, sj);

  if (libs == 1) {

    /* attack1() can return 0, 1 or 3. */
    result = attack1(si, sj, &xi, &xj, komaster);
    if (result)
      READ_RETURN(read_result, i, j, xi, xj, result);
  } 
  else if (libs == 2) {

    /* attack2() can return 0, 1 or 2. */
    result = attack2(si, sj, &xi, &xj, komaster);
    if (result)
      READ_RETURN(read_result, i, j, xi, xj, result);

  } else if (libs == 3) {

    result =  attack3(si, sj, &xi, &xj, komaster);
    if (result)
      READ_RETURN(read_result, i, j, xi, xj, result);

  } else if ((stackp <= fourlib_depth) && (libs == 4)) {
    result =  attack4(si, sj, &xi, &xj, komaster);
    if (result)
      READ_RETURN(read_result, i, j, xi, xj, result);
  }

  /* More than 4 liberties.  The attack fails. 
   * (This is an oversimplification, but it's the one we use here.)
   */
  READ_RETURN0(read_result);
}


/* If (si, sj) points to a group with exactly one liberty, attack1
 * determines whether it can be captured by playing at this liberty.
 * If yes, (*i, *j) is the killing move. i & j may be NULL if caller is
 * only interested in whether it can be captured.
 *
 * The attack may fail for two different reasons. The first one is
 * that the attack may be an illegal ko capture, in this case 3 is
 * returned (need to play a ko threat before the attack can be
 * fulfilled).
 *
 * The second cause for failure is that the attack is caught in a
 * snapback. We must require that it is a proper snapback, though. By
 * proper snapback we mean a position like
 *
 *  XXXXO
 *  XO.XO
 *  XOXOO
 *  -----
 *
 * where capture by O and recapture by X leaves the X stone intact
 * with at least two liberties:
 *
 *  XXXXO
 *  X..XO
 *  X.XOO
 *  -----
 *
 * There are a number of different kinds of improper snapbacks, which
 * have in common that the attacked string ends up captured. We don't
 * consider these as failures to attack. Three examples are given below.
 *
 *   XXOOOOO     (X can recapture but loses most of the string.)
 *   X.XXXXO
 *   -------
 *
 *   XXXOOOOOOOO (Like the previous example, except O loses one more stone)
 *   XO*XXXXXXXO
 *   -----------
 *
 *   XXXOO       (After three captures, the lone X stone is gone.)
 *   XO.XO
 *   -----
 *
 * This function is fast and never branches. There's little point in
 * caching the result.
 */

static int
attack1(int si, int sj, int *i, int *j, int komaster)
{
  int xi, xj;
  int color = p[si][sj];
  int other = OTHER_COLOR(color);
  int result = -1;
  
  SETUP_TRACE_INFO("attack1", si, sj);
  reading_node_counter++;
  
  /* Pick up the position of the liberty. */
  findlib(si, sj, 1, &xi, &xj);

  /* If the attacked string consists of more than one stone, the
   * attack never fails. (This assumes simple ko rule. With superko
   * rule it could still be a ko violation.)
   */
  if (countstones(si, sj) > 1) {
    result = 1;
  }
  /* Try to play on the liberty. This fails if and only if it is an
   * illegal ko capture.
   */
  else if (trymove(xi, xj, other, "attack1-A", si, sj)) {
    /* If the attacking string is also a single stone, a possible recapture
     * would be a ko violation, not snapback. Hence guaranteed success.
     */
    if (countstones(xi, xj) == 1) {
      result = 1;
    }
    /* Is the attacker in atari? If not the attack was successful. */
    else if (countlib(xi, xj) > 1) {
      result = 1;
    }
    /* It was in atari, do recapture. Notice that the liberty must be
     * at (si, sj) since we have already established that this string
     * was a single stone.
     */
    else if (trymove(si, sj, color, "attack1-B", si, sj)) {
      /* If this was a proper snapback, (si, sj) will now have more
       * than one liberty.
       */
      if (countlib(si, sj) > 1) {
	/* Proper snapback, attack fails. */
	RTRACE("SNAPBACK at %m!\n", i, j);
	result = 0;
      }
      else
	result = 1;
      popgo();
    }
    popgo();
  }
  else {/* Illegal ko capture. */
    if (komaster != color) 
      result = 3;
    else
      result = 0;
  }
  
  if (result > 0) {
    if (i) *i = xi;
    if (j) *j = xj;
    SGFTRACE(xi, xj, result, NULL);
  }
  else {
    SGFTRACE(-1, -1, 0, NULL);
  }
  
  return result;
}


/* If si, sj points to a group with exactly two liberties
 * attack2 determines whether it can be captured in ladder or net.
 * If yes, *i, *j is the killing move. i & j may be null if caller 
 * is only interested in whether it can be captured.
 *  
 * Returns 2 or 3 if it can be killed conditioned on ko. Returns
 * 2 if it can be killed provided (other) is willing to ignore
 * any ko threat. Returns 3 if (other) wins provided he has a
 * ko threat which must be answered. Can give a return code 3 
 * yet (*i,*j)=(-1,-1) if the winning move is an illegal
 * ko capture. In this case, making a ko threat and having it
 * answered should transform the position to one where the return
 * code is 2. 
 *
 * See the comment before defend1 about ladders and reading depth.  */

static int 
attack2(int si, int sj, int *i, int *j, int komaster)
{
  int color, other;
  int ai, aj, bi, bj;
  int di, dj, gi, gj, hi, hj;
  int xi, xj;
  int liberties, r, can_save;
  int libi[3], libj[3];
  int adj, adji[MAXCHAIN], adjj[MAXCHAIN], adjsize[MAXCHAIN], adjlib[MAXCHAIN];
  int savei = -1, savej = -1;
  int savecode=0;
  int val;
  int             found_read_result;
  Read_result   * read_result;

  SETUP_TRACE_INFO("attack2", si, sj);
  reading_node_counter++;

  find_origin(si, sj, &si, &sj);
  assert(p[si][sj]!=EMPTY);
  assert(countlib(si, sj) == 2);

  RTRACE("checking attack on %m with 2 liberties\n", si, sj);

  color=p[si][sj];
  other=OTHER_COLOR(color);

  if ((stackp <= depth) && (hashflags & HASH_ATTACK2)) {
  
    found_read_result = get_read_result(ATTACK2, &si, &sj, &read_result);
    if (found_read_result) {
      TRACE_CACHED_RESULT(*read_result);
      if (rr_get_result(*read_result) != 0) {
	if (i) *i = rr_get_result_i(*read_result);
	if (j) *j = rr_get_result_j(*read_result);
      }

      SGFTRACE(rr_get_result_i(*read_result), rr_get_result_j(*read_result),
	       rr_get_result(*read_result), "cached");
      return rr_get_result(*read_result);
    }

    /* This data should always be recorded. */
    if (read_result) {
      rr_set_routine_i_j_stackp(*read_result, ATTACK2, si, sj, stackp);
    }
  } else
    read_result = NULL;

  /* The attack may fail if a boundary string is in atari and cannot 
   * be defended.  First we must try defending such a string. 
   *
   * We start by trying to defend the boundary string by looking for an
   * adjacent string which is in atari. If capturing that string at (gi, gj)
   * doesn't work, we then try to extend the boundary string at (adji[r],
   * adji[r]) to (hi, hj). If that doesn't work, then we try to find a
   * boundary string to (adji[r], adjj[r]) which is in atari, and we try
   * capturing at (gi, gj).  
   */
  chainlinks(si, sj, &adj, adji, adjj, adjsize, adjlib);
  for (r = 0; r < adj; r++) {
    if (adjlib[r]!=1) 
      continue;

    /* if stackp > depth and any boundary chain is in atari, assume safe.
     * However, if the captured chain is only of size 1, there can still
     * be a working ladder, so continue if that is the case.
     */
    if (stackp > depth && adjsize[r] > 1) {
      SGFTRACE(-1, -1, 0, "boundary in atari");
      READ_RETURN0(read_result);
    }
    
    if (relative_break_chain(adji[r], adjj[r], &gi, &gj, si, sj)) {
      if (is_ko(gi, gj, other)) {
	savecode = 2;
	savei = gi;
	savej = gj;
      }
      else {
	SGFTRACE(gi, gj, 1, "relative break chain");
	READ_RETURN(read_result, i, j, gi, gj, 1);
      }
    }

    findlib(adji[r], adjj[r], 1, &hi, &hj);
    if (trymove(hi, hj, other, "attack2-A", si, sj)) {
      int dcode = do_find_defense(si, sj, NULL, NULL, komaster);
      if ((dcode != 1) 
	  && (do_attack(si, sj, NULL, NULL, komaster)))
	{
	  if (dcode == 0) {
	    popgo();
	    SGFTRACE(hi, hj, 1, "saving boundary");
	    READ_RETURN(read_result, i, j, hi, hj, 1);
	  }
	  else if (dcode == 2)
	    savecode = 3;
	  else if (dcode == 3)
	    savecode = 2;
	  savei = hi;
	  savej = hj;
	}
      popgo();
    }
    else if (is_ko(hi, hj, other) 
	     && (stackp <= ko_depth)
	     && (komaster != color)
	     && (tryko(hi, hj, other, "attack2-B"))) {
      if (do_attack(si, sj, NULL, NULL, other)) {
	savecode = 3;
	/* Since (hi,hj) is an illegal ko capture we return pass. */
	savei = -1;
	savej = -1;
      }
    }
  }

  /* Get the two liberties of (si, sj) into (ai, aj) and (bi, bj). */ 
  liberties = findlib(si, sj, 2, libi, libj);
  ASSERT(liberties == 2, si, sj);
  order_moves(si, sj, liberties, libi, libj, other);

  ai = libi[0];
  aj = libj[0];
  bi = libi[1];
  bj = libj[1];

  /* FIXME: Ununroll this loop. */
  RTRACE("considering atari at %m\n", ai, aj);

  /* we only want to consider the move at (ai,aj) if:
   * stackp <= backfill_depth
   * -or-  stackp <= depth and it is an isolated stone
   * -or-  it is not in immediate atari
   */

  if ((stackp <= backfill_depth) 
      || ((stackp <= depth) 
	  && ((ai == 0) || (p[ai-1][aj] != other))
	  && ((ai == board_size-1) || (p[ai+1][aj] != other))
	  && ((aj == 0) || (p[ai][aj-1] != other))
	  && ((aj == board_size-1) || (p[ai][aj+1] != other)))
      || !self_atari(ai, aj, other)) {
    if (trymove(ai, aj, other, "attack2-B", si, sj)) {
      can_save = defend1(si, sj, &di, &dj, komaster);
      if (can_save == 0) {
	popgo();
	RTRACE("%m captures !!\n", ai, aj);
	SGFTRACE(ai, aj, 1, "atari - A");
	READ_RETURN(read_result, i, j, ai, aj, 1);
      } else if (can_save == 2) {
	savei = ai;
	savej = aj;
	savecode = 3;
      } else if (can_save == 3) {
	savei = ai;
	savej = aj;
	savecode = 2;
      }
      popgo();
    }
    else if (is_ko(ai, aj, other)
	     && (komaster != color)
	     && tryko(ai, aj, other, "attack2-D")) {
      can_save = defend1(si, sj, &di, &dj, other);
      if (can_save != 1) {
	savei = ai;
	savej = aj;
	savecode = 3;
      }
      popgo();
    }
  }

  /* Try backfilling if atari is impossible. */
  if ((stackp <= backfill_depth)
      && approxlib(ai, aj, other, 2, libi, libj) == 1) {
    int ui = libi[0];
    int uj = libj[0];
    if (trymove(ui, uj, other, "attack2-C", si, sj))  {
      if (do_attack(si, sj, NULL, NULL, komaster) 
	  && !do_find_defense(si, sj, NULL, NULL, komaster)) {
	savei = ui;
	savej = uj;
	savecode = 1;
      }
      popgo();
    } else if (is_ko(ui, uj, other) && (savecode == 0)) {
      savei = ai;
      savej = aj;
      savecode = 3;
    }
  }
  
  RTRACE("first atari didn't work - try %m\n", bi, bj);
  
  /* we only want to consider the move at (bi,bj) if:
   * stackp <= backfill_depth
   * -or-  stackp <= depth and it is an isolated stone
   * -or-  it is not in immediate atari
   */
  
  if ((stackp <= backfill_depth) 
      || ((stackp <= depth) 
	  && ((bi == 0) || (p[bi-1][bj] != other))
	  && ((bi == board_size-1) || (p[bi+1][bj] != other))
	  && ((bj == 0) || (p[bi][bj-1] != other))
	  && ((bj == board_size-1) || (p[bi][bj+1] != other)))
      || !self_atari(bi, bj, other)) {
    if (trymove(bi, bj, other, "attack2-D", si, sj)) {
      can_save = defend1(si, sj, &di, &dj, komaster);
      if (can_save == 0) {
	RTRACE("%m captures !!\n", bi, bj);
	popgo();
	SGFTRACE(bi, bj, 1, "atari - B");
	READ_RETURN(read_result, i, j, bi, bj, 1);
      }
      else if ((can_save == 2) && (savecode == 0)) {
	savei = bi;
	savej = bj;
	savecode = 3;
      } if ((can_save == 3) 
	    && ((savecode == 0) || (savecode == 3))) {
	savei = bi;
	savej = bj;
	savecode = 2;
      }
      popgo();
    }
    else if (is_ko(bi, bj, other) 
	     && (komaster != color)
	     && tryko(bi, bj, other, "attack2-G")) {
      can_save = defend1(si, sj, &di, &dj, other);
      if (can_save != 1) {
	savei = bi;
	savej = bj;
	savecode = 3;
      }
      popgo();
    }
  }
  
  /* Try backfilling if atari is impossible. */
  if ((stackp <= backfill_depth)
      && (approxlib(bi, bj, other, 2, libi, libj) == 1)) {
    int ui = libi[0];
    int uj = libj[0];
    if (trymove(ui, uj, other, "attack2-E", si, sj)) {
      if (do_attack(si, sj, NULL, NULL, komaster) 
	  && !do_find_defense(si, sj, NULL, NULL, komaster)) {
	savei = ui;
	savej = uj;
	savecode = 1;
      }
      popgo();
    }
  } else
    RTRACE("second atari at %m illegal\n", bi, bj);

  /* The simple ataris didn't work. Try something more fancy. */
  if (find_cap2(si, sj, &xi, &xj, komaster)) {
    SGFTRACE(xi, xj, 1, "find cap2");
    READ_RETURN(read_result, i, j, xi, xj, 1);
  }

  if (stackp <= backfill_depth) {
    val = special_attack2(si, sj, libi, libj, &xi, &xj, komaster);
    if (val == 1) {
      SGFTRACE(xi, xj, 1, "special attack2");
      READ_RETURN(read_result, i, j, xi, xj, 1);
    }
    else if (val > 1) {
      savei = xi;
      savej = xj;
      savecode = val;
    }
  }

  if (savei == -1) {
    RTRACE("ALIVE!!\n");
    SGFTRACE(-1, -1, 0, NULL);
    READ_RETURN0(read_result);
  }

  SGFTRACE(savei, savej, savecode, "saved move");
  READ_RETURN(read_result, i, j, savei, savej, savecode);
}


/* attack3(si, sj, *i, *j) is used when (si, sj) points to a group with
 * three liberties. It returns true if it finds a way to kill the group.
 *
 * Return code is 2 if the group can be killed if the attacker is 
 * willing to ignore any ko threat.
 *
 * Return code is 3 if the group can be killed if the attacker is 
 * able to find a ko threat which must be answered.
 *
 * If non-NULL (*i, *j) will be set to the move which makes the
 * attack succeed.
 */

static int 
attack3(int si, int sj, int *i, int *j, int komaster)
{
  int color = p[si][sj];
  int other = OTHER_COLOR(color);
  int adj, adji[MAXCHAIN], adjj[MAXCHAIN], adjsize[MAXCHAIN], adjlib[MAXCHAIN];
  int xi, xj;
  int liberties;
  int libi[3], libj[3];
  int r;
  int dcode = 0;
  int k;
  int mx[MAX_BOARD][MAX_BOARD];    /* mark moves already tried */
  int             found_read_result;
  int result = 0;
  Read_result   * read_result;

  SETUP_TRACE_INFO("attack3", si, sj);
  reading_node_counter++;

  assert(p[si][sj] != EMPTY);

  if ((stackp <= depth) && (hashflags & HASH_ATTACK3)) {
    found_read_result = get_read_result(ATTACK3, &si, &sj, &read_result);
    if (found_read_result) {
      TRACE_CACHED_RESULT(*read_result);
      if (rr_get_result(*read_result) != 0) {
	if (i) *i = rr_get_result_i(*read_result);
	if (j) *j = rr_get_result_j(*read_result);
      }

      SGFTRACE(rr_get_result_i(*read_result), rr_get_result_j(*read_result),
	       rr_get_result(*read_result), "cached");
      return rr_get_result(*read_result);
    }

    /* This data should always be recorded. */
    if (read_result) {
      rr_set_routine_i_j_stackp(*read_result, ATTACK3, si, sj, stackp);
    }
  } else
    read_result = NULL;

  if (stackp > depth) {
    SGFTRACE(-1, -1, 0, "stackp > depth");
    READ_RETURN0(read_result);
  }

  memset(mx, 0, sizeof(mx));

  chainlinks(si, sj, &adj, adji, adjj, adjsize, adjlib);
  for (r=0; r<adj; r++) {
    if (adjlib[r] == 1) {
      findlib(adji[r], adjj[r], 1, &xi, &xj);
      
      /* If a component of the surrounding chain has just
       * one liberty, we try to defend it.
       */

      if (!mx[xi][xj] && !self_atari(xi, xj, other)) {
	if (trymove(xi, xj, other, "attack3-A", si, sj)) {
	  mx[xi][xj] = 1;
	  dcode = do_find_defense(si, sj, NULL, NULL, komaster);
	  if ((dcode != 1) && do_attack(si, sj, NULL, NULL, komaster))
	    {
	      if (dcode == 0) 
		result = 1;
	      else if (dcode == 2)
		result = 3;
	      else if (dcode == 3)
		result = 2;
	      popgo();
	      SGFTRACE(xi, xj, result, "defending chain - A");
	      READ_RETURN(read_result, i, j, xi, xj, result);
	    }
	  popgo();
	}
      }

      /* Try to rescue the boundary piece by capturing. */
      result = relative_break_chain(adji[r], adjj[r], &xi, &xj, si, sj);
      if (result) {
	SGFTRACE(xi, xj, result, "relative break chain");
	READ_RETURN(read_result, i, j, xi, xj, result);
      }
    }
    else if (adjlib[r] == 2 && (stackp <= fourlib_depth)) {
      if (attack2(adji[r], adjj[r], NULL, NULL, komaster) 
	  && find_defense(adji[r], adjj[r], &xi, &xj)) {
	if (trymove(xi, xj, other, "attack3-B", si, sj)) {
	  if (do_attack(si, sj, NULL, NULL, komaster)) {
	    dcode = find_defense(si, sj, NULL, NULL);
	    if (dcode != 1) {
	      if (dcode == 0) 
		result = 1;
	      else if (dcode == 2)
		result = 3;
	      else if (dcode == 3)
		result = 2;
	      popgo();
	      SGFTRACE(xi, xj, result, "defending chain - B");
	      READ_RETURN(read_result, i, j, xi, xj, result);
	    }
	  }
	  popgo();
	}
      }
    }
  }

  RTRACE("checking attack on %m with 3 liberties\n", si, sj);
  liberties = findlib(si, sj, 3, libi, libj);
  ASSERT (liberties == 3, si, sj);
  for (k=0; k < liberties; k++)
    if (edge_attack(si, sj, libi[k], libj[k], &xi, &xj))
      READ_RETURN(read_result, i, j, xi, xj, result);
  order_moves(si, sj, liberties, libi, libj, other);

  for (k = 0; k < liberties; ++k) {
    if (stackp >= branch_depth && k > 0)
      break;
    xi = libi[k];
    xj = libj[k];
    
    if (!mx[xi][xj] && trymove(xi, xj, other, "attack3-C", si, sj)) {
      mx[xi][xj] = 1;
      RTRACE("try attacking at %m ...\n", xi, xj);
      dcode = find_defense(si, sj, NULL, NULL);
      if ((dcode != 1) && do_attack(si, sj, NULL, NULL, komaster)) {
	if (dcode == 0)
	  result = 1;
	else if (dcode == 2)
	  result = 3;
	else if (dcode == 3)
	  result = 2;
	popgo();
	SGFTRACE(xi, xj, result, "attacking at liberty");
	READ_RETURN(read_result, i, j, xi, xj, result);
      }
      popgo();
    }
  }

  /* The simple ataris didn't work. Try something more fancy. */
  if ((stackp < backfill_depth) && find_cap3(si, sj, &xi, &xj, komaster)) {
    SGFTRACE(xi, xj, 1, "find cap3");
    READ_RETURN(read_result, i, j, xi, xj, 1);
  }

  SGFTRACE(-1, -1, 0, NULL);
  READ_RETURN0(read_result);
}


/* attack4 tries to capture a string with 4 liberties. This function
 * is not cached.
 */

static int 
attack4(int si, int sj, int *i, int *j, int komaster)
{
  int color = p[si][sj];
  int other = OTHER_COLOR(color);
  int ai, aj;
  int xi, xj;
  int r;
  int k;
  int liberties;
  int libi[4], libj[4];
  int mx[MAX_BOARD][MAX_BOARD];  
  int adj, adji[MAXCHAIN], adjj[MAXCHAIN], adjsize[MAXCHAIN], adjlib[MAXCHAIN];

  SETUP_TRACE_INFO("attack4", si, sj);
  
  assert(p[si][sj]!=EMPTY);
  reading_node_counter++;
  
  if (stackp > depth) {
    SGFTRACE(-1, -1, 0, "stackp > depth");
    return 0;
  }
  
 /* The attack may fail if a boundary string is in atari and cannot be
  * defended. First we must try defending such a string.
  */
  
  memset(mx, 0, sizeof(mx));

  chainlinks(si, sj, &adj, adji, adjj, adjsize, adjlib);
  for (r=0; r<adj; r++) {
    if (adjlib[r] == 1) {
      findlib(adji[r], adjj[r], 1, &xi, &xj);

      /* If moving out results in more than one liberty,
         we resume the attack. */
      if (!mx[xi][xj] && !self_atari(xi, xj, other)) {
	if (trymove(xi, xj, other, "attack4-A", si, sj)) {
	  mx[xi][xj] = 1;
	  if (do_attack(si, sj, NULL, NULL, komaster) 
	      && !find_defense(si, sj, NULL, NULL)) {
	    popgo();
	    SGFTRACE(xi, xj, 1, "defend boundary");
	    if (i) *i = xi;
	    if (j) *j = xj;
	    return 1;
	  }
	  popgo();
	}
      }
      if (relative_break_chain(adji[r], adjj[r], &xi, &xj, si, sj)) {
	SGFTRACE(xi, xj, 1, "relative break chain");
	if (i) *i = xi;
	if (j) *j = xj;
	return 1;
      }
    }
  }

  RTRACE("checking attack on %m with 4 liberties\n", si, sj);
  liberties = findlib(si, sj, 4, libi, libj);
  ASSERT (liberties == 4, si, sj);
  order_moves(si, sj, liberties, libi, libj, other);

  /* If find_defense was fully reliable for strings with 3 liberties
   * we would not need the inner test (if !attack( ...,x)
   * ||!does_defend(x, ...)).  Currently it seems that the
   * performance penalty of making defend3 try attacking boundary
   * string with 3 liberties is too great. Instead, we try here to
   * attack the stone just played, and if it cannot be captured, or if
   * attacking it does not defend (si,sj), the attack succeeds.
   */

  for (k = 0; k < liberties; ++k) {
    if (stackp >= branch_depth && k > 0)
      break;
    ai = libi[k];
    aj = libj[k];

    if (trymove(ai, aj, other, "attack4-B", si, sj)) {
      RTRACE("try attacking at %m ...\n", ai, aj);
      ASSERT(countlib(si, sj) == 3, si, sj);
      if (!find_defense(si, sj, NULL, NULL)
	  && do_attack(si, sj, NULL, NULL, komaster)) {
	if (!do_attack(ai, aj, &xi, &xj, komaster) ||
	    !does_defend(xi, xj, si, sj)) {
	  popgo();
	  SGFTRACE(ai, aj, 1, "attack just played stone");
	  if (i) *i = ai;
	  if (j) *j = aj;
	  return 1;
	}
      }
      popgo();
    }
  }

  SGFTRACE(-1, -1, 0, NULL);
  return 0;
}


/* If (si,sj) points to a string with 2 liberties, find_cap2(si,sj,&i,&j)
 * looks for a configuration of the following type:
 *
 *  X.
 *  .*
 *
 * where X is an element of the string in question. It tries the
 * move at * and returns true this move captures the string, leaving
 * (i,j) pointing to *. 
 */

static int
find_cap2(int si, int sj, int *i, int *j, int komaster)
{
  int ai, aj, bi, bj;
  int libi[3], libj[3];
  int ti = -1, tj = -1;

  /* Get the liberties into (ai, aj) and (bi, bj). */
  if (findlib(si, sj, 2, libi, libj) != 2)
    return 0;
  ai = libi[0];
  aj = libj[0];
  bi = libi[1];
  bj = libj[1];

  /* Check if the two liberties are located like the figure above. */
  if (((ai == bi+1) || (ai == bi-1)) 
      && ((aj == bj+1) || (aj == bj-1))) 
  {
    /* Which of the two corner points should we use? One of them is 
     * always occupied by the string at (m, n), the other one is either
     * free or occupied by something else.
     */
    if (p[bi][aj] == EMPTY) {
      ti = bi;
      tj = aj;
    }
    if (p[ai][bj] == EMPTY) {
      ti = ai;
      tj = bj;
    }

    /* If we didn't find a free intersection, we couldn't make the move. */
    if (ti == -1)
      return 0;

    /* Ok, we found the spot. Now see if the move works. */
    RTRACE("trying to capture %m with capping move at %m\n", si, sj, ti, tj);
    if (trymove(ti, tj, OTHER_COLOR(p[si][sj]),"find_cap2", si, sj)) {
      if (defend2(si, sj, NULL, NULL, komaster)) {
	RTRACE("cap failed!\n", si, sj, ti, tj);
	popgo();
	return 0;
      }
      RTRACE("cap succeeded!\n", si, sj, ti, tj);
      popgo();
      if (i) *i = ti;
      if (j) *j = tj;
      return 1;
    }
  }

  return 0;
}    


/* If (si,sj) points to a string with 3 liberties, find_cap3(si,sj,&i,&j)
 * looks for a configuration of the following type:
 *
 *  XXa
 *  cb*
 *
 * where X are elements of the string in question and a, b and c are
 * its liberties. It tries the move at * and returns true this move
 * captures the string, leaving (i,j) pointing to *.
 */

static int
find_cap3(int si, int sj, int *i, int *j, int komaster)
{
  int ai, aj, bi, bj, ci, cj;
  int libi[3], libj[3];
  int ti = -1, tj = -1;
  int k;

  /* Get the liberties into (ai, aj) and (bi, bj). */
  if (findlib(si, sj, 3, libi, libj) != 3)
    return 0;

  for (k = 0; k < 3; ++k) {
    /* k and k+1 mod 3 will be (0,1), (1,2) and (2,0); These are the 
     * three combinations of indices that we have to test the same
     * way that find_cap2 does.  Then after that we see if the third
     * liberty fits into the pattern.
     */
    ai = libi[k];
    aj = libj[k];
    bi = libi[(k+1)%3];
    bj = libj[(k+1)%3];
    ci = libi[(k+2)%3];
    cj = libj[(k+2)%3];

    /* Check if the two liberties are located like the figure above. */
    if (!(((ai == bi+1) || (ai == bi-1)) 
	  && ((aj == bj+1) || (aj == bj-1))))
      continue;

    /* Which of the two corner points should we use? One of them is 
     * always occupied by the string at (m, n), the other one is either
     * free or occupied by something else.
     */
    if (p[bi][aj] == EMPTY) {
      ti = bi;
      tj = aj;
      break;
    }
    if (p[ai][bj] == EMPTY) {
      ti = ai;
      tj = bj;
      break;
    }
  }
  
  /* If we didn't find a free intersection, we couldn't make the move. */
  if (ti == -1)
    return 0;

  /* Now see if the third liberty fits into the pattern.
   * The different combinations are:
   *
   *    1     2    3     4
   *   XXb   XXa   Xc    Xc
   *   ca*   cb*   Xa    Xb
   *               b*    a*
   *
   *   and its mirrors.
   */
  if (!((ti == ai && ci == ai && (cj == aj+(aj-tj)))    /* 1 */
	|| (ti == bi && ci == bi && (cj == bj+(bj-tj))) /* 2 */
	|| (tj == aj && cj == aj && (ci == ai+(ai-ti))) /* 3 */
	|| (tj == bj && cj == bj && (ci == bi+(bi-ti))) /* 4 */
	))
    return 0;

  /* Ok, we found the spot. Now see if the move works. */
  RTRACE("trying to capture %m with capping move at %m\n", si, sj, ti, tj);
  if (trymove(ti, tj, OTHER_COLOR(p[si][sj]),"find_cap3", si, sj)) {
    if (defend3(si, sj, NULL, NULL, komaster)) {
      RTRACE("cap3 failed!\n", si, sj, ti, tj);
      popgo();
      return 0;
    }
    RTRACE("cap3 succeeded!\n", si, sj, ti, tj);
    popgo();
    if (i) *i = ti;
    if (j) *j = tj;
    return 1;
  }

  return 0;
}


/* In a situation like this:
 *       
 * -----        the code that
 * cO.OX        follows can find
 * XXOOX        the attacking move
 * XO.OX        at 'c=(xi,xj)'.
 * XOOOX
 * XXXXX
 *
 * The name of the function corresponds to special_rescue2, which is
 * fairly similar to this situation.
 */

static int
special_attack2(int si, int sj, int libi[2], int libj[2], 
		int *ti, int *tj, int komaster)
{
  int  color=p[si][sj];
  int  other=OTHER_COLOR(color);
  int  savecode = 0;
  int  newlibi[3], newlibj[3];
  int  xi, xj;
  int  savei = -1, savej = -1;
  int  k;


  for (k = 0; k < 2; ++k) {
    if (suicide(libi[k], libj[k], other) 
	&& (approxlib(libi[k], libj[k], color, 3, newlibi, newlibj)==2)) {
      if ((newlibi[0] != libi[1-k]) || (newlibj[0] != libj[1-k])) {
	xi = newlibi[0];
	xj = newlibj[0];
      } else {
	xi = newlibi[1];
	xj = newlibj[1];
      }

      if (!self_atari(xi, xj, other)
	  && trymove(xi, xj, other, "special_attack2", si, sj)) {
	if (do_attack(si, sj, NULL, NULL, komaster)) {
	  int dcode = find_defense(si, sj, NULL, NULL);
	  if (dcode != 1) {
	    if (dcode == 0) {
	      popgo();

	      if (ti) *ti = xi;
	      if (tj) *tj = xj;
	      return 1;
	    }
	    else if (dcode == 2)
	      savecode = 3;
	    else if (dcode == 3)
	      savecode = 2;
	    savei = xi;
	    savej = xj;
	  }
	}
	popgo();
      }
    }
  }

  if (savecode) {
    if (ti) *ti = savei;
    if (tj) *tj = savej;
    return savecode;
  }

  return 0;
}


/* 
 * If (si,sj) points to a string and (ai,aj) a liberty of it,
 * returns 1 if (ai,aj) is an edge liberty in the following
 * configuration:
 * 
 * Ob?
 * Xa.
 * ---
 *
 * and if a move at 'b' captures the X string.
 * 
 * FIXME: (1) So far we do not treat ko;
 *        (2) Sometimes the right move is one space
 *            to the right. We should consider this
 *            if stackp < backfill_depth.
 */

static int
edge_attack(int si, int sj, int ai, int aj, int *ki, int *kj)
{
  int color = p[si][sj];
  int other = OTHER_COLOR(color);

  if ((ai == 0) 
      && (p[ai+1][aj] == EMPTY)
      && ((aj > 0) && (aj < board_size-1))
      && (((p[ai][aj-1] == color)
	   && (p[ai+1][aj-1] == other)
	   && (p[ai][aj+1] == EMPTY))
	  || ((p[ai][aj+1] == color)
	      && (p[ai+1][aj+1] == other)
	      && (p[ai][aj-1] == EMPTY))))
    if (trymove(ai+1, aj, other, "edge_attack", si, sj)) {
      if (!find_defense(si, sj, NULL, NULL)) {
	if (ki) *ki = ai+1;
	if (ki) *kj = aj;
	popgo();
	return 1;
      }
      popgo();
    }
  if ((ai == board_size-1) 
      && (p[ai-1][aj] == EMPTY)
      && ((aj > 0) && (aj < board_size-1))
      && (((p[ai][aj-1] == color)
	   && (p[ai-1][aj-1] == other)
	   && (p[ai][aj+1] == EMPTY))
	  || ((p[ai][aj+1] == color)
	      && (p[ai-1][aj+1] == other)
	      && (p[ai][aj-1] == EMPTY))))
    if (trymove(ai-1, aj, other, "edge_attack", si, sj)) {
      if (!find_defense(si, sj, NULL, NULL)) {
	if (ki) *ki = ai-1;
	if (ki) *kj = aj;
	popgo();
	return 1;
      }
      popgo();
    }
  if ((aj == 0) 
      && (p[ai][aj+1] == EMPTY)
      && ((ai > 0) && (ai < board_size-1))
      && (((p[ai-1][aj] == color)
	   && (p[ai-1][aj+1] == other)
	   && (p[ai+1][aj] == EMPTY))
	  || ((p[ai+1][aj] == color)
	      && (p[ai+1][aj+1] == other)
	      && (p[ai-1][aj] == EMPTY))))
    if (trymove(ai, aj+1, other, "edge_attack", si, sj)) {
      if (!find_defense(si, sj, NULL, NULL)) {
	if (ki) *ki = ai;
	if (ki) *kj = aj+1;
	popgo();
	return 1;
      }
      popgo();
    }
  if ((aj == board_size-1) 
      && (p[ai][aj-1] == EMPTY)
      && ((ai > 0) && (ai < board_size-1))
      && (((p[ai-1][aj] == color)
	   && (p[ai-1][aj-1] == other)
	   && (p[ai+1][aj] == EMPTY))
	  || ((p[ai+1][aj] == color)
	      && (p[ai+1][aj-1] == other)
	      && (p[ai-1][aj] == EMPTY))))
    if (trymove(ai, aj-1, other, "edge_attack", si, sj)) {
      if (!find_defense(si, sj, NULL, NULL)) {
	if (ki) *ki = ai;
	if (ki) *kj = aj-1;
	popgo();
	return 1;
      }
      popgo();
    }
  return 0;
}



/* ================================================================ */
/*            Defending by attacking surrounding strings            */
/* ================================================================ */


/* 
 * (si, sj) points to a string. 
 * 
 * break_chain(si, sj, *i, *j, *k, *l) returns 1 if
 *   part of some surrounding string is in atari, and if capturing
 *   this string results in a live string at (si, sj). 
 *
 * Returns 2 if the capturing string can be taken (as in a snapback),
 *   or the saving move depends on ignoring a ko threat;
 * 
 * Returns 3 if the saving move requires making a ko threat and winning
 *   the ko.
 *
 * (i,j), if not NULL, are left pointing to the appropriate defensive move.
 * (k,l), if not NULL, are left pointing to the boundary string which
 *   is in atari.
 */

static int
break_chain(int si, int sj, int *i, int *j, int *k, int *l, int komaster)
{
  int color, other;
  int r;
  int ai, aj;
  int xi, xj;
  int adj, adji[MAXCHAIN], adjj[MAXCHAIN], adjsize[MAXCHAIN], adjlib[MAXCHAIN];
  int savei = -1, savej = -1;
  int savek = -1, savel = -1;
  int savecode = 0;

  reading_node_counter++;
  
  RTRACE("in break_chain at %m\n", si, sj);
  color = p[si][sj];
  other = OTHER_COLOR(color);

  chainlinks(si, sj, &adj, adji, adjj, adjsize, adjlib);
  if (verbose) {
    RTRACE("chain surrounding %m (liberties): ", si, sj);
    for (r=0; r<adj; r++) {
      ai = adji[r];
      aj = adjj[r];
      RTRACE("%o%m(%d) ", ai, aj, adjlib[r]);  /* %o cancels indent */
    }
    RTRACE("%o\n");
  }

  for (r=0; r<adj; r++) {
    if (adjlib[r] > 1)
      continue;

    ai = adji[r];
    aj = adjj[r];
    if (findlib(ai, aj, 1, &xi, &xj) == 1) {
      if (trymove(xi, xj, color, "break_chain-A", si, sj)) {
	int acode = do_attack(si, sj, NULL, NULL, komaster);
	if (acode != 1) {
	  RTRACE("%m found to defend %m by attacking %m\n", 
		 xi, xj, si, sj, ai, aj);
	  if ((acode != 0) || do_attack(xi, xj, NULL, NULL, komaster)) {
	    savei = xi;
	    savej = xj;
	    savek = ai;
	    savel = aj;
	    if (acode == 0)
	      savecode = 1;
	    else if (acode == 2)
	      savecode = 3;
	    else if (acode == 3)
	      savecode = 2;
	    popgo();
	  }
	  else {
	    if (i) *i = xi;
	    if (j) *j = xj;
	    if (k) *k = ai;
	    if (l) *l = aj;
	    popgo();
	    return 1;
	  }
	}
	else
	  popgo();
      } else {
	if ((stackp <= ko_depth) 
	    && (savecode == 0) 
	    && is_ko(xi, xj, color)) {
	  if ((komaster != other )
	       && tryko(xi, xj, color, "break_chain-B")) {
	    /* (xi,xj) is the appropriate defensive move but it is an illegal
	     * ko capture now so we return pass if no better move found (time
	     * to make ko threat).  
	     */
	    /* FIXME: do_attack never returns -1, does it?
	     *        Check for other occurances of the same mistake.
	     */
	    if (do_attack(si, sj, NULL, NULL, color) != -1) {
	      savei = -1;
	      savej = -1;
	      savek = ai;
	      savel = aj;
	      savecode = 3;
	    }
	    popgo();
	  }
	}
      }
    }
  }

  if (savei != -1) {
    if (i) *i = savei;
    if (j) *j = savej;
    if (k) *k = savek;
    if (l) *l = savel;

    return savecode;
  }

  return 0;
}


/*
 * (si,sj) points to a group. break_chain2(si, sj, *i, *j)
 * returns 1 if there is a string in the surrounding chain having
 * exactly two liberties whose attack leads to the rescue of
 * (si, sj). Then (*i, *j) points to the location of the attacking move.
 * 
 * Returns 2 if the attacking stone can be captured, 1 if it cannot.
 */

static int 
break_chain2(int si, int sj, int *i, int *j, int komaster)
{
  int color, other;
  int r;
  int k;
  int u=0, v;
  int ai, aj;
  int xi, xj;
  int saveti = -1, savetj = -1;
  int adj;
  int adji[MAXCHAIN], adjj[MAXCHAIN], adjsize[MAXCHAIN], adjlib[MAXCHAIN];
  int libi[2], libj[2];
  int liberties;
  int ti[MAXCHAIN], tj[MAXCHAIN];
  int mw[MAX_BOARD][MAX_BOARD];
  int savecode = 0;

  memset(mw, 0, sizeof(mw));

  reading_node_counter++;
  
  RTRACE("in break_chain2 at %m\n", si, sj);
  color = p[si][sj];
  other = OTHER_COLOR(color);
  chainlinks(si, sj, &adj, adji, adjj, adjsize, adjlib);
  RTRACE("chain around %m (liberties): ", si, sj);
  for (r=0;r<adj;r++) {
    ai = adji[r];
    aj = adjj[r];

    /* We make a list in the (adji, adjj) array of the liberties
     * of boundary strings having exactly two liberties. We mark
     * each liberty in the mw array so that we do not list any
     * more than once.
     */
    if (adjlib[r] == 2) {
      findlib(ai, aj, 2, libi, libj);
      for (k = 0; k < 2; ++k) {
	if (!mw[libi[k]][libj[k]] && !liberty_of(libi[k], libj[k], si, sj)) {
	  mw[libi[k]][libj[k]] = 1;
	  ti[u] = libi[k];
	  tj[u] = libj[k];
	  u++;
	}
      }

      if ((stackp <= fourlib_depth)
	  && break_chain(adji[r], adjj[r], &xi, &xj, NULL, NULL, komaster)
	  && !mw[xi][xj]) {
	mw[xi][xj] = 1;
	ti[u] = xi;
	tj[u] = xj;
	u++;
      }
    }
    RTRACE("%o%m(%d) ", ai, aj, adjlib[r]);
  }
  RTRACE("%o\n");

  order_moves(si, sj, u, ti, tj, color);

  /* We do not wish to consider the move if it can be 
   * immediately recaptured, unless stackp <= backfill_depth.
   *
   * With the incremental board code it's not much point in using
   * approxlib() as heuristic whether we want to play the move. Better
   * and more accurate to really play the move and then use findlib() to
   * see how many liberties we obtained.
   */

  for (v=0; v<u; v++) {
    if (!trymove(ti[v], tj[v], color, "break_chain2-A", si, sj))
      continue;
    
    liberties = findlib(ti[v], tj[v], 2, libi, libj);
    if (liberties > 1) {
      int acode = do_attack(si, sj, i, j, komaster);
      if (acode != 1) {
	if (acode == 0) {
	  if (i) *i = ti[v];
	  if (j) *j = tj[v];
	  popgo();
	  return 1;
	}
	else if (acode == 2)
	  savecode = 3;
	else if (acode == 3)
	  savecode = 2;
	saveti = ti[v];
	savetj = tj[v];
      }
    }
    else if (stackp <= backfill_depth) {
      int bi = libi[0];
      int bj = libj[0];
      int try_harder = 0;

      assert(liberties == 1);

      if (trymove(bi, bj, other, "break_chain2-B", si, sj)) {
	if (p[si][sj] != EMPTY) {
	  int dcode = find_defense(si, sj, NULL, NULL);
	  if (dcode) {
	    if (dcode == 1) {
	      if (i) *i=ti[v];
	      if (j) *j=tj[v];
	      popgo();
	      popgo();
	      return 1;
	    }
	    else if ((dcode == 2) || (savecode == 0)) {
	      savecode = dcode;
	      saveti = ti[v];
	      savetj = tj[v];
	    }
	  }
	}
	if (countlib(bi, bj) == 1)
	  try_harder=1;
	popgo();
      }
      if (try_harder) {
	int acode=do_attack(si, sj, i, j, komaster);
	if (acode != 1) {
	  if (acode == 0) {
	    if (i) *i = ti[v];
	    if (j) *j = tj[v];
	    popgo();
	    return 1;
	  }
	  else if (acode == 2)
	    savecode = 3;
	  else if (acode == 3)
	    savecode = 2;
	  saveti = ti[v];
	  savetj = tj[v];
	}
      }
    }
    popgo(); /* (ti[v], tj[v]) */
  }

  if (saveti != -1) {
    if (i) *i = saveti;
    if (j) *j = savetj;
    return savecode;
  }

  return 0;
}


/*
 * (si,sj) points to a group. break_chain3(si, sj, *i, *j)
 * returns 1 if there is a string in the surrounding chain having
 * exactly three liberties whose attack leads to the rescue of
 * (si, sj). Then (*i, *j) points to the location of the attacking move.
 * 
 * Returns 2 if the attacking stone can be captured, 1 if it cannot.
 */

static int 
break_chain3(int si, int sj, int *i, int *j, int komaster)
{
  int color, other;
  int r;
  int k;
  int u=0, v;
  int ai, aj;
  int saveti = -1, savetj = -1;
  int adj;
  int adji[MAXCHAIN], adjj[MAXCHAIN], adjsize[MAXCHAIN], adjlib[MAXCHAIN];
  int libi[3], libj[3];
  int ti[MAXCHAIN], tj[MAXCHAIN];
  int mw[MAX_BOARD][MAX_BOARD];
  int savecode = 0;
  int liberties = countlib(si, sj);

  memset(mw, 0, sizeof(mw));

  reading_node_counter++;
  
  RTRACE("in break_chain3 at %m\n", si, sj);
  color = p[si][sj];
  other = OTHER_COLOR(color);
  chainlinks(si, sj, &adj, adji, adjj, adjsize, adjlib);
  RTRACE("chain around %m (liberties): ", si, sj);
  for (r=0;r<adj;r++) {
    ai = adji[r];
    aj = adjj[r];

    /* We make a list in the (adji, adjj) array of the liberties
     * of boundary strings having exactly three liberties. We mark
     * each liberty in the mw array so that we do not list any
     * more than once.
     */
    if (adjlib[r] == 3) {
      findlib(ai, aj, 3, libi, libj);
      for (k = 0; k < 3; ++k) {
	if (!mw[libi[k]][libj[k]]) {
	  mw[libi[k]][libj[k]] = 1;
	  ti[u] = libi[k];
	  tj[u] = libj[k];
	  u++;
	}
      }

    }
  }
  
  /* We do not wish to consider the move if it can be 
   * immediately recaptured, unless stackp <= backfill_depth.
   *
   * With the incremental board code it's not much point in using
   * approxlib() as heuristic whether we want to play the move. Better
   * and more accurate to really play the move and then use findlib() to
   * see how many liberties we obtained.
   */

  for (v=0; v<u; v++) {
    if (!trymove(ti[v], tj[v], color, "break_chain3-A", si, sj))
      continue;
    
    /* If we just filled our own liberty we back out now */
    if (countlib(si, sj) >= liberties) {
      int acode = do_attack(si, sj, i, j, komaster);
      if (acode != 1) {
	if (acode == 0) {
	  if (i) *i = ti[v];
	  if (j) *j = tj[v];
	  popgo();
	  return 1;
	}
	else if (acode == 2)
	  savecode = 3;
	else if (acode == 3)
	  savecode = 2;
	saveti = ti[v];
	savetj = tj[v];
      }
    }
    popgo(); /* (ti[v], tj[v]) */
  }

  if (saveti != -1) {
    if (i) *i = saveti;
    if (j) *j = savetj;
    return savecode;
  }

  return 0;
}


/*
 * If si, sj points to a group, double_atari_chain2(si, sj, *i, *j)
 * returns 1 if there is move which makes a double atari on 
 * a string in the surrounding chain, and if this move rescues
 * the string. This is a faster, more specific function than
 * break_chain2, and substitutes for it while reading ladders
 * when stackp > depth.
 * 
 * If not NULL, *i, *j points to the location of the attacking move.
 * 
 */

static int 
double_atari_chain2(int si, int sj, int *i, int *j)
{
  int r, k;
  int color = p[si][sj];
  int ai, aj;
  int adj;
  int adji[MAXCHAIN], adjj[MAXCHAIN], adjsize[MAXCHAIN], adjlib[MAXCHAIN];
  int libi[2], libj[2];
  int mw[MAX_BOARD][MAX_BOARD];
  int w = 0;
  int v;
  int ti[MAXCHAIN], tj[MAXCHAIN]; /* list of double atari moves */

  reading_node_counter++;
  
  memset(mw, 0, sizeof(mw));

  color = p[si][sj];
  chainlinks(si, sj, &adj, adji, adjj, adjsize, adjlib);
  RTRACE("chain around %m (liberties): ", si, sj);
  for (r=0; r<adj; r++) {
    ai = adji[r];
    aj = adjj[r];
    if (adjlib[r] == 2) {
      findlib(ai, aj, 2, libi, libj);
      for (k = 0; k < 2; ++k) {
	mw[libi[k]][libj[k]]++;
	if (mw[libi[k]][libj[k]] > 1) {
	  /* found a double atari */
	  ti[w] = libi[k];
	  tj[w] = libj[k];
	  w++;
	}
      }
    }
  }

  for (v=0; v<w; v++)
    if (trymove(ti[v], tj[v], color, "double_atari_chain2", si, sj)) {
      if (!attack(si, sj, NULL, NULL)) {
	popgo();
	if (i) *i = ti[v];
	if (j) *j = tj[v];
	return 1;
      }
      else
	popgo();
    }

  return 0;
}


/* 
 * relative_break_chain(si, sj, *i, *j, ti, tj) is a variant of
 * break_chain. The strings (si, sj) and (ti, tj) are of
 * opposite color, and (ti, tj) is under attack. This function
 * looks for a boundary string to (si, sj) which is in atari,
 * and asks whether capturing it will result in the capture
 * of (si, sj).
 */

static int
relative_break_chain(int si, int sj, int *i, int *j, int ti, int tj)
{
  int color;
  int r;
  int ai, aj;
  int xi, xj;
  int adj, adji[MAXCHAIN], adjj[MAXCHAIN], adjsize[MAXCHAIN], adjlib[MAXCHAIN];
  
  reading_node_counter++;
  
  if (stackp > depth)
    return 0;

  RTRACE("in relative_break_chain at %m (%m)\n", si, sj, ti, tj);
  color = p[si][sj];
  chainlinks(si, sj, &adj, adji, adjj, adjsize, adjlib);

  if (verbose) {
    RTRACE("chain surrounding %m (liberties): ", si, sj);
    for (r=0; r<adj; r++) {
      ai = adji[r];
      aj = adjj[r];
      RTRACE("%o%m(%d) ", ai, aj, adjlib[r]);  /* %o cancels indent */
    }
    RTRACE("%o\n");
  }

  for (r=0; r<adj; r++) {
    if (adjlib[r] > 1)
      continue;

    ai = adji[r];
    aj = adjj[r];
    if (findlib(ai, aj, 1, &xi, &xj) == 1) {
      if (trymove(xi, xj, color, "relative_break_chain", si, sj)) {
	if (attack(ti, tj, NULL, NULL) 
	    && !find_defense(ti, tj, NULL, NULL)) {
	  RTRACE("%m found to attack %m by defending %m\n", 
		 xi, xj, ti, tj, si, sj);
	  if (i) *i = xi;
	  if (j) *j = xj;
	  popgo();
	  return 1;
	}
	else
	  popgo();
      }
    }
  }

  return 0;
}


/* ================================================================ */
/*                          Prioritization                          */
/* ================================================================ */


/* The string at (si, sj) is under attack. The num_moves moves in
 * (movei[], movej[]) for color color have been deemed interesting in
 * the attack or defense of the group. Most of these moves will be
 * immediate liberties of the group.
 *
 * This function orders the moves in the order where the move most 
 * likely to succeed to defend the string will be first and so on.
 *
 * Currently, this is defined as:
 * 1) Moves which let the defending string get more liberties are more
 *    interesting.
 * 2) Moves adjacent to the most open liberties are more 
 *    interesting than those with fewer open liberties.
 * 3) Moves on the edge are less interesting.
 */

static void
order_moves(int si, int sj, int num_moves, int *movei, int *movej, int color)
{
  int  origini, originj;
  int  scores[MAX_BOARD * MAX_BOARD];
  int  move;
  int  i, j;
  int  maxscore;
  int  max_at;
  int  string_color = p[si][sj];
  int  k;

  find_origin(si, sj, &origini, &originj);

  /* Each move will have a score. First, we clear the score array. */
  memset(scores, 0, num_moves * sizeof(int));

  /* Assign a score to each move. */
  for (move = 0; move < num_moves; ++move) {
    int mi = movei[move];
    int mj = movej[move];

    /* Look at the neighbors of this move and count the things we
     * find. Friendly and opponent stones are related to color, i.e.
     * the player to move, not to the color of the string.
     *
     * If this piece of code should turn out to be a bottleneck it can
     * be speeded up by loop unrolling and/or moving it into the
     * incremental board code.
     */
    int number_edges       = 0; /* outside board */
    int number_same_string = 0; /* the string being attacked */
    int number_own         = 0; /* friendly stone */
    int number_opponent    = 0; /* opponent stone */
    int captured_stones    = 0; /* number of stones captured by this move*/
    int number_open        = 0; /* empty intersection */
    
    for (k=0; k<4; k++) {
      i = mi + deltai[k];
      j = mj + deltaj[k];
      if (!ON_BOARD(i, j))
	number_edges++;
      else if (p[i][j] == EMPTY)
	number_open++;
      else {
	if (p[i][j] == string_color) {
	  int ai, aj;
	  find_origin(i, j, &ai, &aj);
	  if (ai == origini && aj == originj)
	    number_same_string++;
	}
	
	if (p[i][j] == color)
	  number_own++;
	else {
	  number_opponent++;
	  if (countlib(i, j) == 1)
	    captured_stones += countstones(i, j);
	}
      }
    }
    
    /* 1) Add twice the number of liberties the group receives by
     *    extending to the intersection of the move. Only
     *    applicable if the move is adjacent to the group.
     */
    if (number_same_string > 0) {
      scores[move] += 2 * approxlib(mi, mj, string_color, 4, NULL, NULL);
    }

    /* 2) If the group is attacked, add the number of liberties the
     *    attacker gets when playing there, but never more than four.
     */
    if (color != string_color) {
      int libs = approxlib(mi, mj, color, 4, NULL, NULL);
      if (libs > 4)
	libs = 4;
      scores[move] += libs;
    }
    
    /* 3) Add the number of open liberties near the move to its score.
     */
    scores[move] += number_open;

    /* 4) Add 5 if the move is not on the edge. 
     */
    if ((mi != 0) 
	&& (mi != board_size-1)
	&& (mj != 0)
	&& (mj != board_size-1))
    scores[move] += 5;
  }

  /* Now sort the moves.  We use selection sort since this array will
   * probably never be more than 10 moves long.  In this case, the
   * overhead imposed by quicksort will probably overshadow the gains
   * given by the O(n*log(n)) behaviour over the O(n^2) behaviour of
   * selection sort.
   */
  for (i = 0; i < num_moves-1; ++i) {

    /* Find the move with the biggest score. */
    maxscore = scores[i];
    max_at = i;
    for (j = i+1; j < num_moves; ++j) {
      if (scores[j] > maxscore) {
	maxscore = scores[j];
	max_at = j;
      }
    }

    /* Now exchange the move at i with the move at max_at.
     * Don't forget to exchange the scores as well.
     */
    if (max_at != i) {
      int tempi = movei[i];
      int tempj = movej[i];
      int tempmax = scores[i];

      movei[i] = movei[max_at];
      movej[i] = movej[max_at];
      scores[i] = scores[max_at];

      movei[max_at] = tempi;
      movej[max_at] = tempj;
      scores[max_at] = tempmax;
    }
  }
}


/* ================================================================ */
/*                         Reading utilities                        */
/* ================================================================ */


static int safe_move_cache[MAX_BOARD][MAX_BOARD][2];
static int safe_move_cache_when[MAX_BOARD][MAX_BOARD][2];
static void clear_safe_move_cache(void);

static void
clear_safe_move_cache(void)
{
  int i,j;

  for (i=0;i<MAX_BOARD;i++)
    for (j=0;j<MAX_BOARD;j++) {
      safe_move_cache_when[i][j][0] = -1;
      safe_move_cache_when[i][j][1] = -1;
    }
}

/* safe_move(i, j, color) checks whether a move at (i, j) is illegal
 * or can immediately be captured. If stackp==0 the result is cached.
 * If the move only can be captured by a ko, it's considered safe.
 * This may or may not be a good convention.
 *
 * For performance reasons, the result of this function is cached.
 */

int 
safe_move(int i, int j, int color)
{
  int safe = 0;
  static int initialized = 0;

  if (!initialized) {
    clear_safe_move_cache();
    initialized = 1;
  }

  /* If we have this position cached, use the previous value. */
  if (stackp == 0
      && safe_move_cache_when[i][j][color==BLACK] == position_number)
    return safe_move_cache[i][j][color==BLACK];

  /* Otherwise calculate the value... */
  if (trymove(i, j, color, "safe_move", -1, -1)) {
    int acode = attack(i, j, NULL, NULL);
    if (acode != 1)
      safe = 1;
    popgo();
  }
  
  /* ...and store it in the cache.
   * FIXME: Only store result in cache when we're working at full depth.
   */
  if (stackp == 0) {
    if (0)
      gprintf("Safe move at %m for %s cached when depth=%d, position number=%d\n",
	      i, j, color_to_string(color), depth, position_number);
    safe_move_cache_when[i][j][color==BLACK] = position_number;
    safe_move_cache[i][j][color==BLACK] = safe;
  }

  return safe;
}


/* ===================== Statistics  ============================= */


/* Clear statistics. */
void
reset_reading_node_counter()
{
  reading_node_counter = 0;
}


/* Retrieve statistics. */
int
get_reading_node_counter()
{
  return reading_node_counter;
}


/* ================================================================ */
/*              Experimental code, not currently in use.            */
/* ================================================================ */


/* naive_ladder(si, sj, &i, &j) tries to capture a string (si, sj)
 * with exactly two liberties under simplified assumptions, which are
 * adequate in a ladder. The rules are as follows:
 *
 * 1. The attacker is allowed to play at each of the two liberties.
 *    If the move was legal, the string now has exactly one
 *    liberty.
 * 2. Define the last stone of the string to be the stone of the
 *    string adjacent to the last liberty. The defender is allowed to
 *    try the following moves:
 *    - extend the string by playing on the liberty
 *    - try to capture the last stone played by the attacker
 *    - try to capture any stone that was put into atari by the
 *      previous move. It is conjectured that it's sufficient to look
 *      for such stones at a distance two from the last liberty.
 *    We only consider captures that can be done immediately.
 * 3. Depending on the resulting number of liberties of the string, we
 *    value each node as follows:
 *    3 or more liberties: the attack has failed
 *    2 liberties:         recurse
 *    1 liberty:           the attack has succeeded
 *    illegal move for the defender: successful attack
 *    illegal move for the attacker: failed attack
 *
 * Return codes are as usual 0 for failure and 1 for success. If the
 * attack was successful, (*i, *j) contains the attacking move, unless
 * i and j are null pointers.
 *
 * The differences compared to the attack2()/defend1() combination for
 * reading ladders is that this one really always reads them to the
 * very end and that it is faster (because it doesn't call
 * break_chain()). In contrast to attack2() though, this function can
 * only be used for capturing in ladders.
 *
 * FIXME: The ladder capture may depend on ko. Need to add the ko
 *        return codes.
 */

int
naive_ladder(int si, int sj, int *i, int *j)
{
  int color, other;
  int ai, aj, bi, bj;
  int acount = 0, bcount = 0;
  int liberties;
  int libi[2], libj[2];
  
  assert(p[si][sj] != EMPTY);
  assert(countlib(si, sj) == 2);
  DEBUG(DEBUG_READING, "naive_ladder(%m)\n", si, sj);

  RTRACE("checking ladder attack on %m with 2 liberties\n", si, sj);

  color = p[si][sj];
  other = OTHER_COLOR(color);

  /* Get the two liberties of (si, sj) into (ai, aj) and (bi, bj). */ 
  liberties = findlib(si, sj, 2, libi, libj);
  ASSERT(liberties==2, si, sj);
  ai = libi[0];
  aj = libj[0];
  bi = libi[1];
  bj = libj[1];

  /* if (bi, bj) looks more promising we wish to switch the two liberties.
   * We check whether (bi, bj) is adjacent to more open liberties than
   * (ai, aj).
   * FIXME: This is taken from attack2(). Not sure whether it is good here too.
   */

  if ((ai>0) && p[ai-1][aj]==EMPTY)
    acount++;
  if ((ai<board_size-1) && p[ai+1][aj]==EMPTY)
    acount++;
  if ((aj>0) && p[ai][aj-1]==EMPTY)
    acount++;
  if ((aj<board_size-1) && p[ai][aj+1]==EMPTY)
    acount++;
  if ((bi>0) && p[bi-1][bj]==EMPTY)
    bcount++;
  if ((bi<board_size-1) && p[bi+1][bj]==EMPTY)
    bcount++;
  if ((bj>0) && p[bi][bj-1]==EMPTY)
    bcount++;
  if ((bj<board_size-1) && p[bi][bj+1]==EMPTY)
    bcount++;

  if (bcount>acount) {
    ai = libi[1];
    aj = libj[1];
    bi = libi[0];
    bj = libj[0];
  }

  RTRACE("considering atari at %m\n", ai, aj);
  
  if (trymove(ai, aj, other, "naive_ladder-A", si, sj)) {
    if (!naive_ladder_defense(si, sj, ai, aj, bi, bj, color, other)) {
      popgo();
      if (i) *i = ai;
      if (j) *j = aj;
      return 1;
    }
    popgo();
  }
	  
  if (trymove(bi, bj, other, "naive_ladder-B", si, sj)) {
    if (!naive_ladder_defense(si, sj, bi, bj, ai, aj, color, other)) {
      popgo();
      if (i) *i = bi;
      if (j) *j = bj;
      return 1;
    }
    popgo();
  }

  /* Neither move worked. */
  return 0;
}


/* Try to save the one-liberty string (si, sj) from being caught in a
 * ladder. (ai, aj) is the last played attacking stone and (bi, bj) is
 * the last remaining liberty.
 */

static int
naive_ladder_defense(int si, int sj, int ai, int aj, int bi, int bj,
		     int color, int other) {
  int liberties;
  
  /* Try to capture the just played stone. */
  if (naive_ladder_break_through(si, sj, ai, aj, color, other))
    return 1;

  /* Try to run away by extending on the last liberty. */
  if (trymove(bi, bj, color, "naive_ladder_defense", si, sj)) {
    liberties = countlib(si, sj);
    if (liberties >= 3
	|| (liberties == 2
	    && !naive_ladder(si, sj, NULL, NULL))) {
      popgo();
      return 1;
    }
    popgo();
  }
  
  /* Try to capture a string at distance two (Manhattan metric) from
   * the last liberty.
   */
  if (naive_ladder_break_through(si, sj, bi-2, bj, color, other))
    return 1;

  if (naive_ladder_break_through(si, sj, bi-1, bj-1, color, other))
    return 1;
  
  if (naive_ladder_break_through(si, sj, bi, bj-2, color, other))
    return 1;
  
  if (naive_ladder_break_through(si, sj, bi+1, bj-1, color, other))
    return 1;
  
  if (naive_ladder_break_through(si, sj, bi+2, bj, color, other))
    return 1;
  
  if (naive_ladder_break_through(si, sj, bi+1, bj+1, color, other))
    return 1;
  
  if (naive_ladder_break_through(si, sj, bi, bj+2, color, other))
    return 1;
  
  if (naive_ladder_break_through(si, sj, bi-1, bj+1, color, other))
    return 1;
  
  /* Nothing worked. */
  return 0;
}


/* Try to break out of the ladder by capturing (ai, aj). We must first
 * verify that there is an opponent stone there and that it is in
 * atari so we can capture it immediately. After the capture we count
 * liberties for (si, sj) to see if the ladder is decided yet.
 */
static int
naive_ladder_break_through(int si, int sj, int ai, int aj,
			   int color, int other)
{
  int liberties;
  int bi, bj;
  
  if (ai < 0
      || ai >= board_size
      || aj < 0
      || aj >= board_size)
    return 0;

  if (p[ai][aj] != other)
    return 0;
  
  if (findlib(ai, aj, 1, &bi, &bj) != 1)
    return 0;

  if (trymove(bi, bj, color, "naive_ladder_break_through", si, sj)) {
    liberties = countlib(si, sj);
    if (liberties >= 3
	|| (liberties == 2
	    && !naive_ladder(si, sj, NULL, NULL))) {
      popgo();
      return 1;
    }
    popgo();
  }
  
  return 0;
}



/*
 * Local Variables:
 * tab-width: 8
 * c-basic-offset: 2
 * End:
 */
